# Converts a METIS-graph into the DIMACS format
import sys, os


def convert_to_dimacs(filename):
    if not os.path.isfile(filename):
        print("File not found.")
        sys.exit(0)

    number_nodes = 0
    number_edges = 0
    edges_counted = 0
    adjacency = []

    print("Reading the file.")

    with open(filename) as f:
        node = 0
        for line in f:
            args = line.strip().split()
            if node == 0:
                number_nodes = int(args[0])
                number_edges = int(args[1])
                print(
                    "Graph has "
                    + str(number_nodes)
                    + " nodes and "
                    + str(number_edges)
                    + " edges"
                )
                adjacency = [[] for _ in range(0, number_nodes + 1)]
                adjacency[0] = args
            else:
                adjacency[node] = args
                edges_counted += len(args)
            node += 1

    print("Writing new file")

    filepath = os.path.splitext(filename)
    new_file = filepath[0] + ".dimacs"

    with open(new_file, "w") as f:
        node = 0
        for neighbors in adjacency:
            if node == 0:
                spec = " ".join(neighbors)
                f.write("p " + spec)
                f.write("\n")
            else:
                for neighbor_node in neighbors:
                    if neighbor_node != "":
                        f.write("e " + str(node) + " " + neighbor_node)
                        f.write("\n")
                        if str(node) in adjacency[int(neighbor_node)]:
                            adjacency[int(neighbor_node)].remove(str(node))
            node += 1

    print("Finished converting.")


if __name__ == "__main__":
    convert_to_dimacs(sys.argv[1])



class BaseEnv:
    def __init__(
        self,
        name: str,
        mode: str = "train",
        train_path: str = None,
        val_path: str = None,
        test_path: str = None,
        train_batch_size: int = 1,
        val_batch_size: int = 1,
        test_batch_size: int = 1,
        num_workers: int = 0,
        device: str = "cpu"
    ):
        self.name = name
        self.mode = mode
        self.train_path = train_path
        self.val_path = val_path
        self.test_path = test_path
        self.train_dataset = None
        self.val_dataset = None
        self.test_dataset = None
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.val_batch_size = val_batch_size
        self.num_workers = num_workers
        self.device = device
        
    def load_data(self):
        raise NotImplementedError(
            "``load_data`` is required to implemented in subclasses."
        )

    def train_dataloader(self):
        raise NotImplementedError(
            "``train_dataloader`` is required to implemented in subclasses."
        )

    def val_dataloader(self):
        raise NotImplementedError(
            "``val_dataloader`` is required to implemented in subclasses."
        )

    def test_dataloader(self):
        raise NotImplementedError(
            "``test_dataloader`` is required to implemented in subclasses."
        )



import os
import pickle
import pathlib
import numpy as np
import networkx as nx
from tqdm import tqdm
from ml4co_kit.utils.mis_utils import sat_to_mis_graph
from collections import OrderedDict


class MISSolver:
    def __init__(self, solver_type: str = None) -> None:
        self.solver_type = solver_type
        self.weighted = None
        self.time_limit = 60.0
        self.nodes_num = None
        self.node_labels = None
        self.ref_node_labels = None
        self.sel_nodes_num = None
        self.ref_sel_nodes_num = None
        self.edges = None

    def check_sel_nodes_num_not_none(self):
        if self.sel_nodes_num is None:
            message = (
                "``sel_nodes_num`` cannot be None, please use the method "
                "``solve`` to obtain solutions and use the method ``from_gpickle_folder`` to "
                "get ``sel_nodes_num``."
            )
            raise ValueError(message)

    def check_ref_sel_nodes_num_not_none(self):
        if self.ref_sel_nodes_num is None:
            raise ValueError(
                "``ref_sel_nodes_num`` cannot be None, please use KaMIS to obtain it."
            )

    def from_gpickle_folder(
        self,
        folder: str,
        solve_folder: str = None,
        weighted: bool = False,
        ref: bool = False,
    ):
        if solve_folder is None:
            solve_folder = os.path.join(folder, "solve")
        read_label = True if os.path.exists(solve_folder) else False
        files = os.listdir(folder)
        self.nodes_num = list()
        record_node_labels = OrderedDict()
        record_sel_nodes_num = OrderedDict()
        self.edges = list()
        for filename in files:
            # check the file format
            if not filename.endswith(".gpickle"):
                continue

            # read graph data from .gpickle
            file_path = os.path.join(folder, filename)
            with open(file_path, "rb") as f:
                graph = pickle.load(f)
            graph: nx.Graph

            # nodes num
            nodes_num = graph.number_of_nodes()

            # node labels
            if not read_label:
                node_labels = [_[1] for _ in graph.nodes(data="label")]
                if node_labels is not None and node_labels[0] is not None:
                    node_labels = np.array(node_labels, dtype=np.int64)
                else:
                    node_labels = np.zeros(nodes_num, dtype=np.int64)
                    edges = np.array(graph.edges, dtype=np.int64)
            else:
                solve_filename = filename.replace(
                    ".gpickle", f"_{'weighted' if weighted else 'unweighted'}.result"
                )
                solve_file_path = os.path.join(solve_folder, solve_filename)
                with open(solve_file_path, "r") as f:
                    node_labels = [int(_) for _ in f.read().splitlines()]
                node_labels = np.array(node_labels, dtype=np.int64)
                if node_labels.shape[0] != nodes_num:
                    message = (
                        "The number of nodes in the solution does not match that of"
                        "the problem. Please check the solution."
                    )
                    raise ValueError(message)

            # edges
            edges = np.array(graph.edges, dtype=np.int64)
            edges = np.concatenate([edges, edges[:, ::-1]], axis=0)
            self_loop = np.arange(nodes_num).reshape(-1, 1).repeat(2, axis=1)
            edges = np.concatenate([edges, self_loop], axis=0)
            edges = edges.T

            # add to the list/dict
            self.nodes_num.append(nodes_num)
            record_node_labels[filename] = node_labels
            record_sel_nodes_num[filename] = np.count_nonzero(node_labels)
            self.edges.append(edges)

        if ref:
            self.ref_node_labels = record_node_labels
            self.ref_sel_nodes_num = record_sel_nodes_num
        else:
            self.node_labels = record_node_labels
            self.sel_nodes_num = record_sel_nodes_num

    def read_ref_sel_nodes_num_from_txt(self, file_path: str):
        self.ref_sel_nodes_num = OrderedDict()

        # check the file format
        if not file_path.endswith(".txt"):
            raise ValueError("Invalid file format. Expected a ``.txt`` file.")

        # read the data form .txt
        with open(file_path, "r") as file:
            lines = file.readlines()
            for line in lines:
                line = line.strip()
                split_line = line.split(": ")
                filename = split_line[0]
                ref_sel_nodes_num = int(split_line[1])
                self.ref_sel_nodes_num[filename] = ref_sel_nodes_num

    @staticmethod
    def __prepare_graph(g: nx.Graph, weighted=False):
        raise NotImplementedError(
            "``__prepare_graph`` is required to implemented in subclasses."
        )

    def prepare_instances(
        self, instance_directory: pathlib.Path, cache_directory: pathlib.Path
    ):
        raise NotImplementedError(
            "``prepare_instances`` is required to implemented in subclasses."
        )

    def solve(self, src: pathlib.Path, out: pathlib.Path = None):
        raise NotImplementedError(
            "The method ``solve`` is required to implemented in subclasses."
        )

    def evaluate(self, calculate_gap: bool = False):
        self.check_sel_nodes_num_not_none()
        sel_nodes_num = np.array(list(self.sel_nodes_num.values()))

        if not calculate_gap:
            return_dict = {
                "avg_sel_nodes_num": np.average(sel_nodes_num),
            }
            return return_dict

        # ref_sel_nodes_num
        self.check_ref_sel_nodes_num_not_none()
        ref_sel_nodes_num = np.array(list(self.ref_sel_nodes_num.values()))

        # calculate gap
        gaps = list()
        for filename, ref in self.ref_sel_nodes_num.items():
            solved = self.sel_nodes_num[filename]
            gap = (solved - ref) / ref * 100
            gaps.append(gap)
        gaps = np.array(gaps)

        return_dict = {
            "avg_sel_nodes_num": np.average(sel_nodes_num),
            "avg_ref_sel_nodes_num": np.average(ref_sel_nodes_num),
            "avg_gap": np.average(gaps),
        }
        return return_dict



# Converts a DIMACS-graph into the METIS format
import sys, os, re


def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    return [atoi(c) for c in re.split("(\d+)", text)]


def sort_metis(filename):
    if not os.path.isfile(filename):
        print("File not found.")
        sys.exit(0)

    number_nodes = 0
    number_edges = 0
    edges_counted = 0
    adjacency = []

    print("Reading the file.")

    with open(filename) as f:
        node = 0
        for line in f:
            args = line.strip().split()
            if node == 0:
                number_nodes = int(args[0])
                number_edges = int(args[1])
                print(
                    "Graph has "
                    + str(number_nodes)
                    + " nodes and "
                    + str(number_edges)
                    + " edges"
                )
                adjacency = [[] for _ in range(0, number_nodes + 1)]
                adjacency[0] = [str(number_nodes), str(number_edges)]
            else:
                adjacency[node] = args
                edges_counted += len(args)
            node += 1

    if edges_counted < number_edges:
        print("Found less edges than specified")
        sys.exit(0)

    filepath = os.path.splitext(filename)
    new_file = filepath[0] + "-sorted.graph"

    print("Writing new file to " + new_file)

    with open(new_file, "w") as f:
        node = 0
        for neighbors in adjacency:
            if node != 0:
                neighbors.sort(key=natural_keys)
            if not neighbors:
                f.write(" ")
            else:
                f.write(" ".join(neighbors))
            f.write("\n")
            node += 1

    print("Finished converting.")


if __name__ == "__main__":
    sort_metis(sys.argv[1])



import os
from ml4co_kit.utils import download, extract_archive


class VRPLIBOriDataset:
    def __init__(self) -> None:
        self.url = "https://huggingface.co/datasets/ML4CO/VRPLIBOriDataset/resolve/main/vrplib_original.tar.gz?download=true"
        self.md5 = "7329db3858b318b5ceeab7d0d68f646e"
        self.dir = "dataset/vrplib_original/"
        if not os.path.exists("dataset"):
            os.mkdir("dataset")
        if not os.path.exists(self.dir):
            download(filename="dataset/vrplib_original.tar.gz", url=self.url, md5=self.md5)
            extract_archive(archive_path="dataset/vrplib_original.tar.gz", extract_path=self.dir)



import os
import shutil
import subprocess
import tempfile
import warnings

from .problems import LKHProblem


class NoToursException(Exception):
    pass


def lkh_solve(solver="LKH", problem=None, **params):
    if shutil.which(solver) is None:
        message = (
            f"The LKH cannot be found in the path in your python env. "
            "If you are sure that the LKH is installed, "
            "please confirm whether the Conda environment of the terminal "
            "is consistent with the Python environment."
        )
        raise ValueError(message)

    assert ("problem_file" in params) ^ (
        problem is not None
    ), "Specify a problem object *or* a path."

    if "problem_file" in params:
        # annoying, but necessary to get the original problem dimension
        problem = LKHProblem.load(params["problem_file"])

    if not isinstance(problem, LKHProblem):
        warnings.warn(
            "Subclassing LKHProblem is recommended. Proceed at your own risk!"
        )

    if len(problem.depots) > 1:
        warnings.warn("LKH-3 cannot solve multi-depot problems.")

    prob_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
    problem.write(prob_file)
    prob_file.write("\n")
    prob_file.close()
    params["problem_file"] = prob_file.name

    if "tour_file" not in params:
        tour_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
        params["tour_file"] = tour_file.name
        tour_file.close()

    par_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
    par_file.write("SPECIAL\n")
    for k, v in params.items():
        par_file.write(f"{k.upper()} = {v}\n")
    par_file.close()

    try:
        # stdin=DEVNULL for preventing a "Press any key" pause at the end of execution
        subprocess.check_output(
            [solver, par_file.name], stderr=subprocess.STDOUT, stdin=subprocess.DEVNULL
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.output.decode())

    if (
        not os.path.isfile(params["tour_file"])
        or os.stat(params["tour_file"]).st_size == 0
    ):
        raise NoToursException(
            f"{params['tour_file']} does not appear to contain any tours. LKH probably did not find solution."
        )

    # the tour file produced by LKH-3 includes dummy nodes to indicate depots
    # for example, if a problem has DIMENSION=32 (1 depot node + 31 task nodes),
    # the tour file will have a SINGLE tour with DIMENSION=36 (5 depot nodes + 31 task nodes)
    solution = LKHProblem.load(params["tour_file"])
    tour = solution.tours[0]
    # convert this tour to multiple routes
    routes = []
    route = []
    for node in tour:
        if node in problem.depots or node > problem.dimension:
            if len(route) > 0:
                routes.append(route)
            route = []
        else:
            route.append(node)
    routes.append(route)

    os.remove(par_file.name)
    if "prob_file" in locals():
        os.remove(prob_file.name)
    if "tour_file" in locals():
        os.remove(tour_file.name)

    return routes



import os
import time
import numpy as np
import pathlib
from tqdm import tqdm
from typing import Union
from multiprocessing import Pool
from ml4co_kit.solver.cvrp import CVRPSolver, CVRPPyVRPSolver, CVRPLKHSolver


class CVRPDataGenerator:
    def __init__(
        self,
        num_threads: int = 1,
        nodes_num: int = 50,
        data_type: str = "uniform",
        solver: Union[str, CVRPSolver] = "pyvrp",
        train_samples_num: int = 128000,
        val_samples_num: int = 1280,
        test_samples_num: int = 1280,
        save_path: pathlib.Path = "data/cvrp/uniform",
        filename: str = None,
        # special for demand and capacity
        min_demand: int = 1,
        max_demand: int = 10,
        min_capacity: int = 40,
        max_capacity: int = 40,
        # special for gaussian
        gaussian_mean_x: float = 0.0,
        gaussian_mean_y: float = 0.0,
        gaussian_std: float = 1.0,
    ):
        """
        CVRPDataGenerator
        Args:
            num_threads (int, optional):
                The number of threads to generate datasets.
            nodes_num (int, optional):
                The number of nodes.
            data_type (str, optional):
                The data type.
            solver_type (str, optional):
                The solver type.
            train_samples_num (int, optional):
                The number of training samples.
            val_samples_num (int, optional):
                The number of validation samples.
            test_samples_num (int, optional):
                The number of test samples.
            save_path (pathlib.Path, optional):
                The save path.
            filename (str, optional):
                The filename.
            gaussian_mean_x (float, optional):
                The mean of the x-coordinate in Gaussian data generation.
            gaussian_mean_y (float, optional):
                The mean of the y-coordinate in Gaussian data generation.
            gaussian_std (float, optional):
                The standard deviation in Gaussian data generation.
        """
        # record variable data
        self.num_threads = num_threads
        self.nodes_num = nodes_num
        self.data_type = data_type
        self.solver = solver
        self.train_samples_num = train_samples_num
        self.val_samples_num = val_samples_num
        self.test_samples_num = test_samples_num
        self.save_path = save_path
        self.filename = filename
        # special for demand and capacity
        self.min_demand = min_demand
        self.max_demand = max_demand
        self.min_capacity = min_capacity
        self.max_capacity = max_capacity
        # special for gaussian
        self.gaussian_mean_x = gaussian_mean_x
        self.gaussian_mean_y = gaussian_mean_y
        self.gaussian_std = gaussian_std
        # check the input variables
        self.sample_types = ["train", "val", "test"]
        self.check_num_threads()
        self.check_data_type()
        self.check_solver()
        self.get_filename()

    def check_num_threads(self):
        self.samples_num = 0
        for sample_type in self.sample_types:
            self.samples_num += getattr(self, f"{sample_type}_samples_num")
            if self.samples_num % self.num_threads != 0:
                message = "``samples_num`` must be divisible by the number of threads. "
                raise ValueError(message)

    def check_data_type(self):
        generate_func_dict = {
            "uniform": self.generate_uniform,
            "gaussian": self.generate_gaussian,
        }
        supported_data_type = generate_func_dict.keys()
        if self.data_type not in supported_data_type:
            message = (
                f"The input data_type ({self.data_type}) is not a valid type, "
                f"and the generator only supports {supported_data_type}."
            )
            raise ValueError(message)
        self.generate_func = generate_func_dict[self.data_type]

    def check_solver(self):
        # get solver
        if type(self.solver) == str:
            self.solver_type = self.solver
            supported_solver_dict = {
                "pyvrp": CVRPPyVRPSolver,
                "lkh": CVRPLKHSolver
            }
            supported_solver_type = supported_solver_dict.keys()
            if self.solver_type not in supported_solver_type:
                message = (
                    f"The input solver_type ({self.solver_type}) is not a valid type, "
                    f"and the generator only supports {supported_solver_type}."
                )
                raise ValueError(message)
            self.solver = supported_solver_dict[self.solver_type]()
        else:
            self.solver: CVRPSolver
            self.solver_type = self.solver.solver_type

    def get_filename(self):
        self.filename = (
            f"cvrp{self.nodes_num}_{self.data_type}"
            if self.filename is None
            else self.filename
        )
        self.file_save_path = os.path.join(self.save_path, self.filename + ".txt")
        for sample_type in self.sample_types:
            setattr(
                self,
                f"{sample_type}_file_save_path",
                os.path.join(self.save_path, self.filename + f"_{sample_type}.txt"),
            )
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)

    def generate(self):
        with open(self.file_save_path, "w") as f:
            start_time = time.time()
            for _ in tqdm(
                range(self.samples_num // self.num_threads),
                desc=f"Solving CVRP Using {self.solver_type}",
            ):
                batch_depots_coord, batch_nodes_coord= self.generate_func()
                batch_demands = self.generate_demands()
                batch_capacities = self.generate_capacities()
                with Pool(self.num_threads) as p1:
                    tours = p1.starmap(
                        self.solver.solve,
                        [(batch_depots_coord[idx],
                          batch_nodes_coord[idx],
                          batch_demands[idx],
                          batch_capacities[idx])
                         for idx in range(self.num_threads)],
                    )
                # write to txt
                for idx, tour in enumerate(tours):
                    depot = batch_depots_coord[idx]
                    points = batch_nodes_coord[idx]
                    demands = batch_demands[idx]
                    capicity = batch_capacities[idx][0]
                    f.write("depots " + str(" ").join(str(depot_coord) for depot_coord in depot))
                    f.write(" points" + str(" "))
                    f.write(
                        " ".join(
                            str(x) + str(" ") + str(y)
                            for x, y in points
                        )
                    )
                    f.write(" demands " + str(" ").join(str(demand) for demand in demands))
                    f.write(" capacity " + str(capicity))
                    f.write(str(" output "))
                    f.write(str(" ").join(str(node_idx) for node_idx in tour[0]))
                    f.write("\n")
            end_time = time.time() - start_time
            f.close()
            print(
                f"Completed generation of {self.samples_num} samples of CVRP{self.nodes_num}."
            )
            print(f"Total time: {end_time/60:.1f}m")
            print(f"Average time: {end_time/self.samples_num:.1f}s")
        self.devide_file()

    def devide_file(self):
        with open(self.file_save_path, "r") as f:
            data = f.readlines()
        train_end_idx = self.train_samples_num
        val_end_idx = self.train_samples_num + self.val_samples_num
        train_data = data[:train_end_idx]
        val_data = data[train_end_idx:val_end_idx]
        test_data = data[val_end_idx:]
        data = [train_data, val_data, test_data]
        for sample_type, data_content in zip(self.sample_types, data):
            filename = getattr(self, f"{sample_type}_file_save_path")
            with open(filename, "w") as file:
                file.writelines(data_content)

    def generate_demands(self) -> np.ndarray:
        return np.random.randint(
            low=self.min_demand,
            high=self.max_demand,
            size=(self.num_threads, self.nodes_num)
        )
        
    def generate_capacities(self) -> np.ndarray:
        if self.min_capacity == self.max_capacity:
            return np.ones(shape=(self.num_threads, 1)) * self.min_capacity
        return np.random.randint(
            low=self.min_capacity,
            high=self.max_capacity,
            size=(self.num_threads, 1)
        )
    
    def generate_uniform(self) -> np.ndarray:
        depots = np.random.random([self.num_threads, 2])
        points = np.random.random([self.num_threads, self.nodes_num, 2]) 
        return depots, points

    def generate_gaussian(self) -> np.ndarray:
        depots = np.random.normal(
            loc=[self.gaussian_mean_x, self.gaussian_mean_y],
            scale=self.gaussian_std,
            size=(self.num_threads, 2),
        )
        points = np.random.normal(
            loc=[self.gaussian_mean_x, self.gaussian_mean_y],
            scale=self.gaussian_std,
            size=(self.num_threads, self.nodes_num, 2),
        )
        return depots, points



import numpy as np


def tsp_greedy(
    adj_mat: np.ndarray, 
    np_points: np.ndarray, 
    parallel_sampling: int=1, 
    device: str="cpu", 
    **kwargs
):
    raise NotImplementedError


def tsp_2opt(
    np_points: np.ndarray,
    tours: np.ndarray,
    adj_mat: np.ndarray = None,
    device="cpu",
    **kwargs
):
    raise NotImplementedError



import os
import pickle
import pathlib
import random
import networkx as nx
import numpy as np
from tqdm import tqdm
from typing import Union
from ml4co_kit.solver.mis import KaMISSolver, MISGurobi
from ml4co_kit.solver import MISSolver, KaMISSolver, MISGurobi


class MISDataGenerator:
    def __init__(
        self,
        nodes_num_min: int = 700,
        nodes_num_max: int = 800,
        data_type: str = "er",
        solver: Union[str, MISSolver] = "kamis",
        train_samples_num: int = 128000,
        val_samples_num: int = 1280,
        test_samples_num: int = 1280,
        save_path: pathlib.Path = "data/mis/er",
        filename: str = None,
        # args for generate
        mis_weighted: bool = False,
        er_prob: float = 0.5,
        ba_conn_degree: int = 10,
        hk_prob: float = 0.5,
        hk_conn_degree: int = 10,
        ws_prob: float = 0.5,
        ws_ring_neighbors: int = 2,
    ):
        """
        MISDataGenerator
        Args:
            nodes_num_min (int, optional):
                The minimum number of nodes.
            nodes_num_max (int, optional):
                The maximum number of nodes.
            data_type (str, optional):
                The data type. Support: ``erdos_renyi``,``er``,``barabasi_albert``,
                ``ba``,``holme_kim``,``hk``,``watts_strogatz``,``ws``.
            solver_type (str, optional):
                The solver type. Support: ``kamis``,``gurobi``.
            train_samples_num (int, optional):
                The number of training samples.
            val_samples_num (int, optional):
                The number of validation samples.
            test_samples_num (int, optional):
                The number of test samples.
            save_path (pathlib.Path, optional):
                The save path of mis samples/datasets.
            filename (str, optional):
                The filename of mis samples.
            mis_weighted (bool, optional):
                If enabled, generate the weighted MIS problem instead of MIS.
            er_prob (float, optional):
                The probability parameter for Erdos-Renyi graph generation.
            ba_conn_degree (int, optional):
                The connection degree parameter for Barabasi-Albert graph generation.
            hk_prob (float, optional):
                The probability parameter for Hyperbolic graph generation.
            hk_conn_degree (int, optional):
                The connection degree parameter for Hyperbolic graph generation.
            ws_prob (float, optional):
                The probability parameter for Watts-Strogatz graph generation.
            ws_ring_neighbors (int, optional):
                The number of ring neighbors for Watts-Strogatz graph generation.
            kamis_recompile (bool, optional):
                Flag indicating whether to recompile the KAMIS solver.
        """
        # record variable data
        self.nodes_num_min = nodes_num_min
        self.nodes_num_max = nodes_num_max
        self.data_type = data_type
        self.solver = solver
        self.train_samples_num = train_samples_num
        self.val_samples_num = val_samples_num
        self.test_samples_num = test_samples_num
        self.save_path = save_path
        self.filename = filename
        # args for generate
        self.mis_weighted = mis_weighted
        self.er_prob = er_prob
        self.ba_conn_degree = ba_conn_degree
        self.hk_prob = hk_prob
        self.hk_conn_degree = hk_conn_degree
        self.ws_prob = ws_prob
        self.ws_ring_neighbors = ws_ring_neighbors

        # check the input variables
        self.sample_types = ["train", "val", "test"]
        self.check_data_type()
        self.check_solver()
        self.check_save_path()
        self.get_filename()

    def check_data_type(self):
        generate_func_dict = {
            "erdos_renyi": self.generate_erdos_renyi,
            "er": self.generate_erdos_renyi,
            "barabasi_albert": self.generate_barabasi_albert,
            "ba": self.generate_barabasi_albert,
            "holme_kim": self.generate_holme_kim,
            "hk": self.generate_holme_kim,
            "watts_strogatz": self.generate_watts_strogatz,
            "ws": self.generate_watts_strogatz,
        }
        supported_data_type = generate_func_dict.keys()
        if self.data_type not in supported_data_type:
            message = (
                f"The input data type ({self.data_type}) is not a valid type, "
                f"and the generator only supports {supported_data_type}."
            )
            raise ValueError(message)
        self.generate_func = generate_func_dict[self.data_type]

    def check_save_path(self):
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        for sample_type in self.sample_types:
            path = os.path.join(self.save_path, sample_type)
            setattr(self, f"{sample_type}_save_path", path)
            if not os.path.exists(path):
                os.mkdir(path)

    def get_filename(self):
        if self.filename is None:
            self.filename = (
                f"mis_{self.data_type}_{self.nodes_num_min}_{self.nodes_num_max}"
            )

    def check_solver(self):
        # check solver
        if type(self.solver) == str:
            self.solver_type = self.solver
            supported_solver_dict = {"kamis": KaMISSolver, "gurobi": MISGurobi}
            supported_solver_type = supported_solver_dict.keys()
            if self.solver not in supported_solver_type:
                message = (
                    f"The input solver type ({self.solver}) is not a valid type, "
                    f"and the generator only supports {supported_solver_type}."
                )
                raise ValueError(message)
            self.solver = supported_solver_dict[self.solver]()
        else:
            self.solver: MISSolver
            self.solver_type = self.solver.solver_type
        # check weighted
        if self.mis_weighted != self.solver.weighted:
            message = "``mis_weighted`` and ``solver.weighted`` do not match."
            raise ValueError(message)

    def random_weight(self, n, mu=1, sigma=0.1):
        return np.around(np.random.normal(mu, sigma, n)).astype(int).clip(min=0)

    def generate(self):
        for sample_type in self.sample_types:
            samples_num = getattr(self, f"{sample_type}_samples_num")
            for idx in tqdm(
                range(samples_num),
                desc=f"Generate MIS({self.data_type}) {sample_type}_dataset",
            ):
                filename = f"{self.filename}_{idx}"
                graph = self.generate_func()
                if self.mis_weighted:
                    weight_mapping = {
                        vertex: int(weight)
                        for vertex, weight in zip(
                            graph.nodes,
                            self.random_weight(
                                graph.number_of_nodes(), sigma=30, mu=100
                            ),
                        )
                    }
                    nx.set_node_attributes(graph, values=weight_mapping, name="weight")
                output_file = os.path.join(
                    getattr(self, f"{sample_type}_save_path"), f"{filename}.gpickle"
                )
                with open(output_file, "wb") as f:
                    pickle.dump(graph, f, pickle.HIGHEST_PROTOCOL)

    def solve(self):
        for sample_type in self.sample_types:
            folder = getattr(self, f"{sample_type}_save_path")
            self.solver.solve(folder, folder)

    def generate_erdos_renyi(self) -> nx.Graph:
        num_nodes = random.randint(self.nodes_num_min, self.nodes_num_max)
        return nx.erdos_renyi_graph(num_nodes, self.er_prob)

    def generate_barabasi_albert(self) -> nx.Graph:
        num_nodes = random.randint(self.nodes_num_min, self.nodes_num_max)
        return nx.barabasi_albert_graph(num_nodes, min(self.ba_conn_degree, num_nodes))

    def generate_holme_kim(self) -> nx.Graph:
        num_nodes = random.randint(self.nodes_num_min, self.nodes_num_max)
        return nx.powerlaw_cluster_graph(
            num_nodes, min(self.hk_conn_degree, num_nodes), self.hk_prob
        )

    def generate_watts_strogatz(self) -> nx.Graph:
        num_nodes = random.randint(self.nodes_num_min, self.nodes_num_max)
        return nx.watts_strogatz_graph(num_nodes, self.ws_ring_neighbors, self.ws_prob)



import torch
import numpy as np


class TSPDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        data_path: str,
        mode: str = None,
    ):
        self.data_path = data_path
        self.file_lines = open(data_path).read().splitlines()
        if mode is not None:
            self.mode = mode
        else:
            if "train" in self.data_path:
                self.mode = "train"
            elif "val" in self.data_path:
                self.mode = "val"
            else:
                self.mode = "test"

    def __len__(self):
        return len(self.file_lines)

    def get_example(self, idx):
        # Select sample
        line = self.file_lines[idx]
        # Clear leading/trailing characters
        line = line.strip()
        # Extract points
        points = line.split(" output ")[0]
        points = points.split(" ")
        points = np.array(
            [[float(points[i]), float(points[i + 1])] for i in range(0, len(points), 2)]
        )
        # Extract tour
        tour = line.split(" output ")[1]
        tour = tour.split(" ")
        tour = np.array([int(t) for t in tour])
        tour -= 1

        return points, tour

    def __getitem__(self, idx):
        points, tour = self.get_example(idx)

        # Return a densely connected graph
        adj_matrix = np.zeros((points.shape[0], points.shape[0]))
        for i in range(tour.shape[0] - 1):
            adj_matrix[tour[i], tour[i + 1]] = 1
            adj_matrix[tour[i + 1], tour[i]] = 1

        return (
            torch.LongTensor(np.array([idx], dtype=np.int64)),
            torch.from_numpy(points).float(),
            torch.from_numpy(adj_matrix).float(),
            torch.from_numpy(tour).long(),
        )



import os
import sys
import shutil

root_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(root_folder)
from ml4co_kit.utils.file_utils import compress_folder, extract_archive


def test_file_utils():
    # compress .tar.gz
    compress_folder(
        folder="tests/utils_test/file_utils_test",
        compress_path="tests/utils_test/file_utils_test.tar.gz"
    )
    shutil.rmtree("tests/utils_test/file_utils_test")
    
    # extract_archive .tar.gz
    extract_archive(
        archive_path="tests/utils_test/file_utils_test.tar.gz",
        extract_path="tests/utils_test/file_utils_test"
    )
    os.remove("tests/utils_test/file_utils_test.tar.gz")
    
    # compress .zip
    compress_folder(
        folder="tests/utils_test/file_utils_test",
        compress_path="tests/utils_test/file_utils_test.zip"
    )
    shutil.rmtree("tests/utils_test/file_utils_test")
    
    # extract_archive .zip
    extract_archive(
        archive_path="tests/utils_test/file_utils_test.zip",
        extract_path="tests/utils_test/file_utils_test"
    )
    os.remove("tests/utils_test/file_utils_test.zip")


if __name__ == "__main__":
    test_file_utils()



import os
import torch
import torch.distributed
from torch import nn
from typing import Optional, List
from wandb.util import generate_id
from typing import Union, Optional
from pytorch_lightning.trainer import Trainer
from pytorch_lightning.strategies import Strategy, DDPStrategy
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint, TQDMProgressBar
from pytorch_lightning.utilities import rank_zero_info
from pytorch_lightning.loggers import WandbLogger


class Checkpoint(ModelCheckpoint):
    def __init__(
        self,
        dirpath: str = "wandb/checkpoints",
        monitor: str = "val/loss",
        every_n_epochs: int = 1,
        every_n_train_steps=None,
        filename=None,
    ):
        super().__init__(
            dirpath=dirpath,
            monitor=monitor,
            mode="min",
            save_top_k=-1,
            save_last=True,
            every_n_epochs=every_n_epochs,
            every_n_train_steps=every_n_train_steps,
            filename=filename,
            auto_insert_metric_name=False,
        )


class Logger(WandbLogger):
    def __init__(
        self,
        name: str = "wandb",
        project: str = "project",
        entity: Optional[str] = None,
        save_dir: str = "log",
        id: Optional[str] = None,
        resume_id: Optional[str] = None,
    ):
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        if id is None and resume_id is None:
            wandb_id = os.getenv("WANDB_RUN_ID") or generate_id()
        else:
            wandb_id = id if id is not None else resume_id

        super().__init__(
            name=name, project=project, entity=entity, save_dir=save_dir, id=wandb_id
        )


class Trainer(Trainer):
    def __init__(
        self,
        model: nn.Module,
        # logger
        logger: Optional[Logger] = None,
        wandb_logger_name: str = "wandb",
        resume_id: Optional[str] = None,
        # checkpoint
        ckpt_save_path: Optional[str] = None,
        ckpt_monitor: str = "val/loss",
        ckpt_every_n_epochs: int = 1,
        ckpt_every_n_train_steps: Optional[int] = None,
        ckpt_filename: str = None,
        # trainer basic
        accelerator: str = "auto",
        strategy: Union[str, Strategy] = None,
        devices: Union[List[int], str, int] = "auto",
        fp16: bool = False,
        max_epochs: int = 100,
        max_steps: int = -1,
        val_check_interval: Optional[int] = None,
        log_every_n_steps: Optional[int] = 50,
        gradient_clip_val: int = 1,
        inference_mode: bool = False,
        reload_dataloaders_every_n_epochs: int = 0,
        # Disable JIT profiling executor.
        disable_profiling_executor: bool = True,
        # pretrained
        ckpt_path: Optional[str] = None,
        weight_path: Optional[str] = None
    ):
        # logger
        if logger is None:
            self.logger = Logger(name=wandb_logger_name, resume_id=resume_id)
        else:
            self.logger = logger
        
        # checkpoint
        if ckpt_save_path is None:
            self.ckpt_save_path = os.path.join(
                "train_ckpts", self.logger._name, self.logger._id
            )
        self.ckpt_callback = Checkpoint(
            dirpath=self.ckpt_save_path,
            monitor=ckpt_monitor,
            every_n_epochs=ckpt_every_n_epochs,
            every_n_train_steps=ckpt_every_n_train_steps,
            filename="epoch={epoch}-step={step}" if ckpt_filename is None else ckpt_filename,
        )
        
        # learning rate
        self.lr_callback = LearningRateMonitor(logging_interval="step")

        # strategy
        if strategy is None:
            strategy = DDPStrategy(
                static_graph=True,
                find_unused_parameters=True,
                gradient_as_bucket_view=True
            )
            
        # Disable JIT profiling executor
        if disable_profiling_executor:
            try:
                torch._C._jit_set_profiling_executor(False)
                torch._C._jit_set_profiling_mode(False)
            except AttributeError:
                pass
        
        # super
        super().__init__(
            accelerator=accelerator,
            strategy=strategy,
            devices=devices,
            precision=16 if fp16 else 32,
            logger=self.logger,
            callbacks=[
                TQDMProgressBar(refresh_rate=20),
                self.ckpt_callback,
                self.lr_callback,
            ],
            max_epochs=max_epochs,
            max_steps=max_steps,
            check_val_every_n_epoch=1,
            val_check_interval=val_check_interval,
            log_every_n_steps=log_every_n_steps,
            gradient_clip_val=gradient_clip_val,
            inference_mode=inference_mode,
            reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs
        )
        if ckpt_path is not None:
            model.load_from_checkpoint(ckpt_path)
        elif weight_path is not None:
            model.load_state_dict(torch.load(weight_path))

        self.train_model = model

    def model_train(self):
        rank_zero_info(
            f"Logging to {self.logger.save_dir}/{self.logger.name}/{self.logger.version}"
        )
        rank_zero_info(f"checkpoint_callback's dirpath is {self.ckpt_save_path}")
        rank_zero_info(f"{'-' * 100}\n" f"{str(self.train_model)}\n" f"{'-' * 100}\n")
        self.fit(self.train_model)
        self.logger.finalize("success")

    def model_test(self):
        rank_zero_info(
            f"Logging to {self.logger.save_dir}/{self.logger.name}/{self.logger.version}"
        )
        rank_zero_info(f"{'-' * 100}\n" f"{str(self.train_model)}\n" f"{'-' * 100}\n")
        self.test(self.train_model)



import pickle
import networkx as nx
import matplotlib.pyplot as plt
from .utils import get_pos_layer


def draw_mis_problem(
    save_path: str,
    gpickle_path: str,
    undirected: bool = True,
    self_loop: bool = False,
    pos_type: str = "kamada_kawai_layout",
    figsize: tuple = (5, 5),
    node_color: str = "darkblue",
    edge_color: str = "darkblue",
    node_size: int = 50,
):
    # check the save path
    if "." not in save_path:
        save_path += ".png"

    # read data from .gpickle
    with open(gpickle_path, "rb") as f:
        graph = pickle.load(f)

    # check if undirected/self_loop
    graph: nx.Graph
    if undirected:
        graph = graph.to_undirected()
    self_loop_edges = list(nx.selfloop_edges(graph))
    if self_loop:
        graph.add_edges_from(self_loop_edges)
    else:
        graph.remove_edges_from(self_loop_edges)

    # pos
    pos_layer = get_pos_layer(pos_type)
    pos = pos_layer(graph)

    # plt
    figure = plt.figure(figsize=figsize)
    figure.add_subplot(111)
    nx.draw_networkx_nodes(G=graph, pos=pos, node_color=node_color, node_size=node_size)
    nx.draw_networkx_edges(
        G=graph, pos=pos, edgelist=graph.edges, alpha=1, width=1, edge_color=edge_color
    )

    # save
    plt.savefig(save_path)


def draw_mis_solution(
    save_path: str,
    gpickle_path: str,
    result_path: str,
    undirected: bool = True,
    self_loop: bool = False,
    pos_type: str = "kamada_kawai_layout",
    figsize: tuple = (5, 5),
    sel_node_color: str = "orange",
    unsel_node_color: str = "darkblue",
    edge_color: str = "darkblue",
    node_size: int = 50,
):
    # check the save path
    if "." not in save_path:
        save_path += ".png"

    # read data from .gpickle
    with open(gpickle_path, "rb") as f:
        graph = pickle.load(f)

    # read result from .result
    with open(result_path, "r") as f:
        node_labels = [int(_) for _ in f.read().splitlines()]

    # check if undirected/self_loop
    graph: nx.Graph
    if undirected:
        graph = graph.to_undirected()
    self_loop_edges = list(nx.selfloop_edges(graph))
    if self_loop:
        graph.add_edges_from(self_loop_edges)
    else:
        graph.remove_edges_from(self_loop_edges)

    # pos
    pos_layer = get_pos_layer(pos_type)
    pos = pos_layer(graph)

    # plt
    figure = plt.figure(figsize=figsize)
    figure.add_subplot(111)
    colors = [unsel_node_color if bit == 0 else sel_node_color for bit in node_labels]
    nx.draw_networkx_nodes(G=graph, pos=pos, node_color=colors, node_size=node_size)
    nx.draw_networkx_edges(
        G=graph, pos=pos, edgelist=graph.edges, alpha=1, width=1, edge_color=edge_color
    )

    # save
    plt.savefig(save_path)



from ml4co_kit.data.tsp.tsp_uniform import TSPUniformDataset
from ml4co_kit.solver.tsp.base import TSPSolver


class TSPUniformEvaluator:
    def __init__(self) -> None:
        self.dataset = TSPUniformDataset()
        self.supported = self.dataset.supported

    def show_files(self, nodes_num: int):
        return self.supported[nodes_num]

    def evaluate(
        self,
        solver: TSPSolver,
        file_path: str,
        **solver_args,
    ):
        solver.from_txt(file_path)
        solver.solve(**solver_args)
        return solver.evaluate(calculate_gap=True)



import importlib.util

# base
from .data import TSPLIBOriDataset, TSPUniformDataset, TSPLIB4MLDataset, ML4TSPDataset
from .data import SATLIBOriDataset
from .data import VRPLIBOriDataset
from .evaluate import TSPEvaluator, TSPLIBOriEvaluator, TSPLIB4MLEvaluator, TSPUniformEvaluator
from .evaluate import SATLIBEvaluator
from .evaluate import CVRPEvaluator
from .generator import TSPDataGenerator, MISDataGenerator, CVRPDataGenerator
from .solver import TSPSolver, TSPLKHSolver, TSPConcordeSolver
from .solver import MISSolver, KaMISSolver, MISGurobi
from .solver import CVRPSolver, CVRPPyVRPSolver, CVRPLKHSolver
from .utils import download, compress_folder, extract_archive, _get_md5

# expand - matplotlib
found_matplotlib = importlib.util.find_spec("matplotlib")
if found_matplotlib is not None:
    from .draw.tsp import draw_tsp_problem, draw_tsp_solution
    from .draw.mis import draw_mis_problem, draw_mis_solution
    from .draw.cvrp import draw_cvrp_problem, draw_cvrp_solution

# expand - pytorch_lightning
found_pytorch_lightning = importlib.util.find_spec("pytorch_lightning")
if found_pytorch_lightning is not None:
    from .learning.env import BaseEnv
    from .learning.model import BaseModel
    from .learning.train import Checkpoint, Logger, Trainer
    from .learning.utils import to_numpy, to_tensor
    from .learning.utils import check_dim
    from .learning.utils import points_to_distmat, sparse_points


__version__ = "0.0.5"
__author__ = "SJTU-ReThinkLab"



import os
import bz2
import lzma
import gzip
import codecs
import pickle
import itertools
import numpy as np
import networkx as nx
from tqdm import tqdm
from typing import Tuple
from collections import OrderedDict


class FileObject(object):
    def __init__(self, name, mode="r", compression=None):
        self.fp = None
        self.ctype = None
        self.fp_extra = None
        self.open(name, mode=mode, compression=compression)

    def open(self, name, mode="r", compression=None):
        if compression == "use_ext":
            self.get_compression_type(name)
        else:
            self.ctype = compression

        if not self.ctype:
            self.fp = open(name, mode)
        elif self.ctype == "gzip":
            self.fp = gzip.open(name, mode + "t")
        elif self.ctype == "bzip2":
            try:
                self.fp = bz2.open(name, mode + "t")
            except:
                self.fp_extra = bz2.BZ2File(name, mode)
                if mode == "r":
                    self.fp = codecs.getreader("ascii")(self.fp_extra)
                else:
                    self.fp = codecs.getwriter("ascii")(self.fp_extra)
        else:
            self.fp = lzma.open(name, mode=mode + "t")

    def close(self):
        if self.fp:
            self.fp.close()
            self.fp = None

        if self.fp_extra:
            self.fp_extra.close()
            self.fp_extra = None

        self.ctype = None

    def get_compression_type(self, file_name):
        ext = os.path.splitext(file_name)[1]
        if ext == ".gz":
            self.ctype = "gzip"
        elif ext == ".bz2":
            self.ctype = "bzip2"
        elif ext in (".xz", ".lzma"):
            self.ctype = "lzma"
        else:
            self.ctype = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class CNF(object):
    def __init__(
        self,
        from_file = None,
        from_fp = None,
        comment_lead = ["c"],
    ):
        self.nv = 0
        self.clauses = []
        self.comments = []

        if from_file:
            self.from_file(from_file, comment_lead, compressed_with="use_ext")
        elif from_fp:
            self.from_fp(from_fp, comment_lead)

    def __repr__(self):
        """
        State reproducible string representaion of object.
        """
        s = self.to_dimacs().replace("\n", "\\n")
        return f'CNF(from_string="{s}")'

    def from_file(self, fname, comment_lead = ["c"], compressed_with = "use_ext"):
        with FileObject(fname, mode="r", compression=compressed_with) as fobj:
            self.from_fp(fobj.fp, comment_lead)

    def from_fp(self, file_pointer, comment_lead=["c"]):
        self.nv = 0
        self.clauses = []
        self.comments = []
        comment_lead = set(["p"]).union(set(comment_lead))

        for line in file_pointer:
            line = line.rstrip()
            if line:
                if line[0] not in comment_lead:
                    self.clauses.append(list(map(int, line.split()[:-1])))
                elif not line.startswith("p cnf "):
                    self.comments.append(line)

        self.nv = max(
            map(
                lambda cl: max(map(abs, cl)),
                itertools.chain.from_iterable([[[self.nv]], self.clauses]),
            )
        )


def sat_to_mis_graph(sat_path: str) -> Tuple[nx.Graph, int]:
    cnf = CNF(sat_path)
    nv = cnf.nv
    clauses_num = len(cnf.clauses)
    clauses = list(filter(lambda x: x, cnf.clauses))
    ind = {k: [] for k in np.concatenate([np.arange(1, nv + 1), -np.arange(1, nv + 1)])}
    edges = []
    for i, clause in enumerate(clauses):
        a = clause[0]
        b = clause[1]
        c = clause[2]
        aa = 3 * i + 0
        bb = 3 * i + 1
        cc = 3 * i + 2
        ind[a].append(aa)
        ind[b].append(bb)
        ind[c].append(cc)
        edges.append((aa, bb))
        edges.append((aa, cc))
        edges.append((bb, cc))
    for i in np.arange(1, nv + 1):
        for u in ind[i]:
            for v in ind[-i]:
                edges.append((u, v))
    graph = nx.from_edgelist(edges)
    return graph, clauses_num


def cnf_to_gpickle(file_path: str, save_path: str = None) -> Tuple[nx.Graph, int]:
    # check the file format
    if not file_path.endswith(".cnf"):
        raise ValueError("Invalid file format. Expected a ``.cnf`` file.")
    if save_path is not None:
        if not save_path.endswith(".gpickle"):
            raise ValueError("Invalid file format. Expected a ``.gpickle`` file.")
    
    # convert sat to mis graph
    graph, clauses_num = sat_to_mis_graph(sat_path=file_path)
    if save_path is not None:
        if not save_path.endswith(".gpickle"):
            save_path += ".gpickle"
        with open(save_path, "wb") as f:
            pickle.dump(graph, f, pickle.HIGHEST_PROTOCOL)
    return graph, clauses_num
        

def cnf_folder_to_gpickle_folder(cnf_folder: str, gpickle_foler: str):
    # check the folder
    mis_graph_save_dir = os.path.join(gpickle_foler, "mis_graph")
    if not os.path.exists(mis_graph_save_dir):
        os.makedirs(mis_graph_save_dir)
        
    # cnf to gpickle
    ref_dict = OrderedDict()
    cnf_files = os.listdir(cnf_folder)
    for cnf_file in tqdm(cnf_files, desc=f"Processing files in {cnf_folder}"):
        file_path = os.path.join(cnf_folder, cnf_file)
        gpickle_file = cnf_file.replace(".cnf", ".gpickle")
        save_path = os.path.join(mis_graph_save_dir, gpickle_file)
        _, clauses_num = cnf_to_gpickle(file_path, save_path)
        ref_dict[gpickle_file] = clauses_num
    
    # save the ref. solution
    ref_save_path = os.path.join(gpickle_foler, "ref_solution.txt")
    with open(ref_save_path, 'w') as ref_file:
        for gpickle_file, clauses_num in ref_dict.items():
            ref_file.write(f"{gpickle_file}: {clauses_num}\n")


import torch
import numpy as np
from typing import Union, Tuple
from sklearn.neighbors import KDTree


#####################################################
#                    Check Utils                    #
#####################################################

def check_dim(array: Union[np.ndarray, torch.Tensor], dim: int):
    if type(array) == np.ndarray:
        if array.ndim != dim:
            raise ValueError("Dimension mismatch!")
    if type(array) == torch.Tensor:
        if array.ndim != dim:
            raise ValueError("Dimension mismatch!")
        

#####################################################
#                     Type Utils                    #
#####################################################

def to_numpy(
    x: Union[np.ndarray, torch.Tensor, list]
) -> np.ndarray:
    if type(x) == np.ndarray:
        return x
    elif type(x) == torch.Tensor:
        return x.cpu().detach().numpy()
    elif type(x) == list:
        return np.array(x)
    
    
def to_tensor(
    x: Union[np.ndarray, torch.Tensor, list]
) -> torch.Tensor:
    if type(x) == torch.Tensor:
        return x
    elif type(x) == np.ndarray:
        return torch.from_numpy(x)
    elif type(x) == list:
        return torch.tensor(x)
    

#####################################################
#                    Graph Utils                    #
#####################################################

def sparse_points(
    points: Union[np.ndarray, torch.Tensor],
    sparse_factor: int,
    device: str = "cpu"
) -> Tuple[torch.Tensor, torch.Tensor]:
    points = to_numpy(points)
    if points.ndim == 2:
        points = np.expand_dims(points, axis=0)
    
    edge_index = list()
    for idx in range(points.shape[0]):
        kdt = KDTree(points[idx], leaf_size=30, metric='euclidean')
        _, idx_knn = kdt.query(points[idx], k=sparse_factor, return_distance=True)
        _edge_index_0 = torch.arange(points[idx].shape[0])
        _edge_index_0 = _edge_index_0.reshape((-1, 1))
        _edge_index_0 = _edge_index_0.repeat(1, sparse_factor)
        _edge_index_0 = _edge_index_0.reshape(-1)
        _edge_index_1 = torch.from_numpy(idx_knn.reshape(-1))
        _edge_index = torch.stack([_edge_index_0, _edge_index_1], dim=0)
        edge_index.append(_edge_index.unsqueeze(dim=0))
    edge_index = torch.cat(edge_index, dim=0).to(device)

    points = torch.from_numpy(points).to(device)
    return points, edge_index


def points_to_distmat(
    points: torch.Tensor, edge_index: torch.Tensor = None
) -> torch.Tensor:
    device = points.device
    if edge_index is None:
        if points.ndim == 2:
            distmat = torch.cdist(points, points)
        else:
            distmat = torch.zeros(size=(points.shape[0], points.shape[1], points.shape[1]))
            for i, matrix in enumerate(points):
                distmat[i] = torch.cdist(matrix, matrix)
    else:     
        if points.ndim == 2:
            x = edge_index[0]
            y = edge_index[1]
            points_x = points[x]
            points_y = points[y]
            distmat = torch.norm(points_x - points_y, dim=1)
        else:
            matrix_list = list()
            for i in range(points.shape[0]):
                x = edge_index[i][0]
                y = edge_index[i][1]
                points_x = points[i][x]
                points_y = points[i][y]
                matrix_list.append(torch.norm(points_x - points_y, dim=1))
            distmat = torch.stack(matrix_list)
    return distmat.to(device)



# Converts a DIMACS-graph into the METIS format
import sys, os, re


def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    return [atoi(c) for c in re.split("(\d+)", text)]


def convert_to_metis(filename):
    if not os.path.isfile(filename):
        print("File not found.")
        sys.exit(0)

    number_nodes = 0
    number_edges = 0
    edges_counted = 0
    adjacency = []

    with open(filename) as f:
        for line in f:
            args = line.strip().split()
            if "cf" in args or "co" in args or "c" in args:
                continue
            if len(args) == 4:
                type, source, target, omit = args
            else:
                type, source, target = args
            if type == "p":
                number_nodes = source
                number_edges = target
                print(
                    "Graph has "
                    + number_nodes
                    + " nodes and "
                    + number_edges
                    + " edges"
                )
                adjacency = [[] for _ in range(0, int(number_nodes) + 1)]
            elif type == "e" or type == "a":
                edge_added = False
                if not target in adjacency[int(source)]:
                    adjacency[int(source)].append(target)
                    edge_added = True
                if not source in adjacency[int(target)]:
                    adjacency[int(target)].append(source)
                    edge_added = True
                if edge_added:
                    edges_counted += 1
            else:
                print("Could not read line.")

    adjacency[0].append(number_nodes)
    # adjacency[0].append(number_edges)
    adjacency[0].append(str(edges_counted))

    print("Writing new file")

    filepath = os.path.splitext(filename)
    new_file = filepath[0] + "-sorted.graph"

    with open(new_file, "w") as f:
        node = 0
        for neighbors in adjacency:
            if node != 0:
                neighbors.sort(key=natural_keys)
            if not neighbors:
                f.write(" ")
            else:
                f.write(" ".join(neighbors))
            f.write("\n")
            node += 1

    print("Finished converting.")


if __name__ == "__main__":
    convert_to_metis(sys.argv[1])



import os
from ml4co_kit.utils import download, extract_archive


#############################################
#                  RESOLVE                  #
#############################################

# ATT (only one instance)
TSPLIB_RESOLVE_ATT_PATH = "dataset/tsplib_original/resolved/ATT/problem"
TSPLIB_RESOLVE_ATT_SOLUTION = "dataset/tsplib_original/resolved/ATT/solution"
TSPLIB_RESOLVE_ATT = ["att48"]

# EUC_2D (17 instances)
TSPLIB_RESOLVE_EUC_2D_PATH = "dataset/tsplib_original/resolved/EUC_2D/problem"
TSPLIB_RESOLVE_EUC_2D_SOLUTION = "dataset/tsplib_original/resolved/EUC_2D/solution"
TSPLIB_RESOLVE_EUC_2D = [
    "eil51",
    "berlin52",
    "st70",
    "eil76",
    "pr76",
    "kroA100",
    "kroC100",
    "kroD100",
    "rd100",
    "eil101",
    "lin105",
    "ch130",
    "ch150",
    "tsp225",
    "a280",
    "pr1002",
    "pr2392",
]

# EXXPLICIT (8 instances)
TSPLIB_RESOLVE_EXPLICIT_PATH = "dataset/tsplib_original/resolved/EXPLICIT/problem"
TSPLIB_RESOLVE_EXPLICIT_SOLUTION = "dataset/tsplib_original/resolved/EXPLICIT/solution"
TSPLIB_RESOLVE_EXPLICIT = ["gr24", "fri26", "bayg29", "bays29", "gr48", "gr120", "brg180", "pa561"]

# GEO (5 instances)
TSPLIB_RESOLVE_GEO_PATH = "dataset/tsplib_original/resolved/GEO/problem"
TSPLIB_RESOLVE_GEO_SOLUTION = "dataset/tsplib_original/resolved/GEO/solution"
TSPLIB_RESOLVE_GEO = ["ulysses16", "ulysses22", "gr96", "gr202", "gr666"]


###################################################
#                    UNRESOLVE                    #
###################################################

# ATT (only one instance)
TSPLIB_UNRESOLVE_ATT_PATH = "dataset/tsplib_original/unresolved/ATT"
TSPLIB_UNRESOLVE_ATT = ["att532"]

# CEIL_2D (4 instanceds)
TSPLIB_UNRESOLVE_CEIL_2D_PATH = "dataset/tsplib_original/unresolved/CEIL_2D"
TSPLIB_UNRESOLVE_CEIL_2D = ["dsj1000", "pla7397", "pla33810", "pla85900"]

# EUC_2D (61 instances)
TSPLIB_UNRESOLVE_EUC_2D_PATH = "dataset/tsplib_original/unresolved/EUC_2D"
TSPLIB_UNRESOLVE_EUC_2D = [
    "rat99",
    "kroB100",
    "kroE100",
    "pr107",
    "pr124",
    "bier127",
    "pr136",
    "pr144",
    "kroA150",
    "kroB150",
    "pr152",
    "u159",
    "rat195",
    "d198",
    "kroA200",
    "kroB200",
    "ts225",
    "pr226",
    "gil262",
    "pr264",
    "pr299",
    "lin318",
    "linhp318",
    "rd400",
    "fl417",
    "pr439",
    "pcb442",
    "d493",
    "u574",
    "rat575",
    "p654",
    "d657",
    "u724",
    "rat783",
    "u1060",
    "vm1084",
    "pcb1173",
    "d1291",
    "rl1304",
    "rl1323",
    "nrw1379",
    "fl1400",
    "u1432",
    "fl1577",
    "d1655",
    "vm1748",
    "u1817",
    "rl1889",
    "d2103",
    "u2152",
    "u2319",
    "pcb3038",
    "fl3795",
    "fnl4461",
    "rl5915",
    "rl5934",
    "rl11849",
    "usa13509",
    "brd14051",
    "d15112",
    "d18512"
]

# EXPLICIT (6 instances)
TSPLIB_UNRESOLVE_EXPLICIT_PATH = "dataset/tsplib_original/unresolved/EXPLICIT"
TSPLIB_UNRESOLVE_EXPLICIT = ["gr17", "gr21", "dantzig42", "swiss42", "hk48", "brazil58"]

# GEO (2 instances)
TSPLIB_UNRESOLVE_GEO_PATH = "dataset/tsplib_original/unresolved/GEO"
TSPLIB_UNRESOLVE_GEO = ["burma14", "ali535"]


###################################################
#                 TSPLIBOriDataset                #
###################################################

class TSPLIBOriDataset:
    def __init__(self):
        self.url = "https://huggingface.co/datasets/ML4CO/TSPLIBOriDataset/resolve/main/tsplib_original.tar.gz?download=true"
        self.md5 = "a25f78ef610b6c4ff1cde27f9d5fa6f9"
        self.dir = "dataset/tsplib_original"
        self.raw_data_path = "dataset/tsplib_original.tar.gz"
        if not os.path.exists("dataset"):
            os.mkdir("dataset")
        if not os.path.exists(self.dir):
            download(filename=self.raw_data_path, url=self.url, md5=self.md5)
            extract_archive(archive_path=self.raw_data_path, extract_path=self.dir)

    @property
    def support(self):
        return {
            "resolved": {
                "ATT": {
                    "path": TSPLIB_RESOLVE_ATT_PATH,
                    "solution": TSPLIB_RESOLVE_ATT_SOLUTION,
                    "problem": TSPLIB_RESOLVE_ATT,
                },
                "EUC_2D": {
                    "path": TSPLIB_RESOLVE_EUC_2D_PATH,
                    "solution": TSPLIB_RESOLVE_EUC_2D_SOLUTION,
                    "problem": TSPLIB_RESOLVE_EUC_2D,
                },
                "EXPLICIT": {
                    "path": TSPLIB_RESOLVE_EXPLICIT_PATH,
                    "solution": TSPLIB_RESOLVE_EXPLICIT_SOLUTION,
                    "problem": TSPLIB_RESOLVE_EXPLICIT,
                },
                "GEO": {
                    "path": TSPLIB_RESOLVE_GEO_PATH,
                    "solution": TSPLIB_RESOLVE_GEO_SOLUTION,
                    "problem": TSPLIB_RESOLVE_GEO,
                },
            },
            "unresolved": {
                "ATT": {
                    "path": TSPLIB_UNRESOLVE_ATT_PATH,
                    "problem": TSPLIB_UNRESOLVE_ATT,
                },
                "EUC_2D": {
                    "path": TSPLIB_UNRESOLVE_EUC_2D_PATH,
                    "problem": TSPLIB_UNRESOLVE_EUC_2D,
                },
                "EXPLICIT": {
                    "path": TSPLIB_UNRESOLVE_EXPLICIT_PATH,
                    "problem": TSPLIB_UNRESOLVE_EXPLICIT,
                },
                "GEO": {
                    "path": TSPLIB_UNRESOLVE_GEO_PATH,
                    "problem": TSPLIB_UNRESOLVE_GEO,
                },
            },
        }



import numpy as np
import matplotlib.pyplot as plt
from typing import Union


def draw_cvrp_problem(
    save_path: str,
    depots: Union[list, np.ndarray],
    points: Union[list, np.ndarray],
    figsize: tuple = (8, 8),
    node_size: int = 50
):
    # check
    if "." not in save_path:
        save_path += ".png"
    if type(depots) == list:
        depots = np.array(depots)
    if depots.ndim == 2 and depots.shape[0] == 1:
        depots = depots[0]
    if depots.ndim != 1:
        raise ValueError("the dim of the depots must 1.")
    if type(points) == list:
        points = np.array(points)
    if points.ndim == 3 and points.shape[0] == 1:
        points = points[0]
    if points.ndim != 2:
        raise ValueError("the dim of the points must 2.")
    
    # plot
    _, ax = plt.subplots(figsize=figsize)
    kwargs = dict(c="tab:red", marker="*", zorder=3, s=500)
    ax.scatter(depots[0], depots[1], label="Depot", **kwargs)
    ax.scatter(points[:, 0], points[:, 1], s=node_size, label="Clients")
    ax.grid(color="grey", linestyle="solid", linewidth=0.2)
    ax.set_aspect("equal", "datalim")
    ax.legend(frameon=False, ncol=2)

    # save
    plt.savefig(save_path)
    
    
def draw_cvrp_solution(
    save_path: str,
    depots: Union[list, np.ndarray],
    points: Union[list, np.ndarray],
    tour: Union[list, np.ndarray],
    figsize: tuple = (8, 8),
    node_size: int = 50
):
    # check
    if "." not in save_path:
        save_path += ".png"
    if type(depots) == list:
        depots = np.array(depots)
    if depots.ndim == 2 and depots.shape[0] == 1:
        depots = depots[0]
    if depots.ndim != 1:
        raise ValueError("the dim of the depots must 1.")
    if type(points) == list:
        points = np.array(points)
    if points.ndim == 3 and points.shape[0] == 1:
        points = points[0]
    if points.ndim != 2:
        raise ValueError("the dim of the points must 2.")
    if type(tour) == list:
        tour = np.array(tour)
    if tour.ndim == 2 and tour.shape[0] == 1:
        tour = tour[0]
    if tour.ndim != 1:
        raise ValueError("the dim of the tour must 1.")
    
    # plot
    _, ax = plt.subplots(figsize=figsize)
    kwargs = dict(c="tab:red", marker="*", zorder=3, s=500)
    ax.scatter(depots[0], depots[1], label="Depot", **kwargs)

    coords = np.concatenate([np.expand_dims(depots, axis=0), points], axis=0)
    x_coords = coords[:, 0]
    y_coords = coords[:, 1]
    split_tours = np.split(tour, np.where(tour == 0)[0])[1: -1]
    idx = 0
    for part_tour in split_tours:
        route = part_tour[1:]
        x = x_coords[route]
        y = y_coords[route]

        # Coordinates of clients served by this route.
        if len(route) == 1:
            ax.scatter(x, y, label=f"Route {idx}", zorder=3, s=node_size)
        ax.plot(x, y)
        arrowprops = dict(arrowstyle='->', linewidth=0.25, color='grey')
        ax.annotate(
            text='', 
            xy=(x_coords[0], y_coords[0]), 
            xytext=(x[0], y[0]), 
            arrowprops=arrowprops
        )
        ax.annotate(
            text='', 
            xy=(x[-1], y[-1]), 
            xytext=(x_coords[0], y_coords[0]), 
            arrowprops=arrowprops
        )
    
    ax.set_aspect("equal", "datalim")
    ax.legend(frameon=False, ncol=2)

    # save
    plt.savefig(save_path)


import os
import sys
import shutil

root_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(root_folder)
from ml4co_kit.data import SATLIBOriDataset, VRPLIBOriDataset
from ml4co_kit.evaluate import (
    TSPLIBOriEvaluator, TSPUniformEvaluator, SATLIBEvaluator, TSPLIB4MLEvaluator
)
from ml4co_kit.solver import TSPConcordeSolver, KaMISSolver


def test_tsplib_original_eval():
    eva = TSPLIBOriEvaluator()
    con_solver = TSPConcordeSolver(scale=1)
    result = eva.evaluate(con_solver, norm="EUC_2D")
    gap_avg = result["gaps"][-1]
    if gap_avg >= 1e-2:
        message = (
            f"The average gap ({gap_avg}) of TSPLIB(EUC_2D) solved by TSPConcordeSolver "
            "is larger than or equal to 1e-2%."
        )
        raise ValueError(message)
    eva.evaluate(con_solver, norm="GEO")
    gap_avg = result["gaps"][-1]
    if gap_avg >= 1e-2:
        message = (
            f"The average gap ({gap_avg}) of TSPLIB(GEO) solved by TSPConcordeSolver "
            "is larger than or equal to 1e-2%."
        )
        raise ValueError(message)


def test_tsplib4ml_eval():
    eva = TSPLIB4MLEvaluator()
    con_solver = TSPConcordeSolver()
    result = eva.evaluate(
        con_solver, normalize=True,
        min_nodes_num=50, max_nodes_num=200
    )
    gap_avg = result["gaps"][-1]
    if gap_avg >= 1e-2:
        message = (
            f"The average gap ({gap_avg}) of TSPLIB4ML(normalize=True) solved by TSPConcordeSolver "
            "is larger than or equal to 1e-2%."
        )
        raise ValueError(message)


def test_tsp_uniform_eval():
    eva = TSPUniformEvaluator()
    supported_files = eva.show_files(nodes_num=50)
    test_file_path = supported_files[-1]
    con_solver = TSPConcordeSolver()
    _, _, gap_avg, _ = eva.evaluate(
        solver=con_solver, 
        file_path=test_file_path,
        num_threads=2,
        show_time=True
    )
    print(f"TSPConcordeSolver Gap: {gap_avg}")
    if gap_avg >= 1e-3:
        message = (
            f"The average gap ({gap_avg}) of TSP50 solved by TSPConcordeSolver "
            "is larger than or equal to 1e-3%."
        )
        raise ValueError(message)


def test_satlib_original_eval():
    SATLIBOriDataset()
    test_folder_full = "dataset/satlib_original/test_files"
    test_folder_part = "dataset/satlib_original/test_files_part"
    os.mkdir(test_folder_part)
    sel_files = os.listdir(test_folder_full)[:10]
    for filename in sel_files:
        src = os.path.join(test_folder_full, filename)
        dst = os.path.join(test_folder_part, filename)
        shutil.copy(src=src, dst=dst)
    eva = SATLIBEvaluator(test_folder="dataset/satlib_original/test_files_part")
    solver = KaMISSolver(time_limit=10)
    gap_avg = eva.evaluate(solver)["avg_gap"]
    print(f"KaMISSolver Gap: {gap_avg}")
    if gap_avg >= 0.1:
        message = (
            f"The average gap ({gap_avg}) of MIS solved by KaMISSolver "
            "is larger than or equal to 0.1%."
        )
        raise ValueError(message)


def test_vrplib_original_eval():
    VRPLIBOriDataset()
    

if __name__ == "__main__":
    test_tsplib_original_eval()
    test_tsplib4ml_eval()
    test_tsp_uniform_eval()
    test_satlib_original_eval()
    test_vrplib_original_eval()



import re
import time
import json
import shutil
import subprocess
import numpy as np
import networkx as nx
import os.path
import pathlib
from tqdm import tqdm
from pathlib import Path
from typing import Union
from .base import MISSolver


class KaMISSolver(MISSolver):
    def __init__(
        self,
        weighted: bool = False,
        time_limit: float = 60.0,
    ):
        """
        KaMIS
        Args:
            weighted (bool, optional):
                If enabled, solve the weighted MIS problem instead of MIS.
            time_limit (float, optional):
                Time limit in seconds.
        """
        super(KaMISSolver, self).__init__()
        self.solver_type = "kamis"
        self.weighted = weighted
        self.time_limit = time_limit
        self.kamis_path = pathlib.Path(__file__).parent

    @staticmethod
    def __prepare_graph(g: nx.Graph, weighted=False):
        g.remove_edges_from(nx.selfloop_edges(g))
        n = g.number_of_nodes()
        m = g.number_of_edges()
        wt = 0 if not weighted else 10
        res = f"{n} {m} {wt}\n"
        for n, nbrsdict in g.adjacency():
            line = []
            if weighted:
                line.append(g.nodes(data="weight", default=1)[n])
            for nbr, _ in sorted(nbrsdict.items()):
                line.append(nbr + 1)
            res += " ".join(map(str, line)) + "\n"
        return res

    def prepare_instances(
        self, instance_directory: pathlib.Path, cache_directory: pathlib.Path
    ):
        instance_directory = Path(instance_directory)
        cache_directory = Path(cache_directory)
        for graph_path in instance_directory.rglob("*.gpickle"):
            self.prepare_instance(graph_path.resolve(), cache_directory)

    def prepare_instance(
        self,
        source_instance_file: Union[str, pathlib.Path],
        cache_directory: Union[str, pathlib.Path],
    ):
        source_instance_file = Path(source_instance_file)
        cache_directory = Path(cache_directory)
        cache_directory.mkdir(parents=True, exist_ok=True)
        dest_path = cache_directory / (
            source_instance_file.stem
            + f"_{'weighted' if self.weighted else 'unweighted'}.graph"
        )
        if os.path.exists(dest_path):
            source_mtime = os.path.getmtime(source_instance_file)
            last_updated = os.path.getmtime(dest_path)
            if source_mtime <= last_updated:
                return
        print(f"Updated graph file: {source_instance_file}.")
        g = nx.read_gpickle(source_instance_file)
        graph = KaMISSolver.__prepare_graph(g, weighted=self.weighted)
        with open(dest_path, "w") as res_file:
            res_file.write(graph)

    def solve(
        self, src: Union[str, pathlib.Path], out: Union[str, pathlib.Path] = None
    ):
        message = (
            "Please check KaMIS compilation. " 
            "you can try ``self.recompile_kamis()``. "
            "If you are sure that the KaMIS is correct, "
            "please confirm whether the Conda environment of the terminal"
            "is consistent with the Python environment."
        )    

        src = Path(src)
        if out is None:
            out = src / "solve"
        else:
            out = Path(out)
        try:
            self._solve(src, out)
        except TypeError:
            raise TypeError(message)
        except FileNotFoundError:
            raise FileNotFoundError(message)

    def _solve(self, src: Union[str, pathlib.Path], out: Union[str, pathlib.Path]):
        src = Path(src)
        out = Path(out)
        if not os.path.exists(out):
            os.makedirs(out)
        cache_directory = src / "preprocessed"
        self.prepare_instances(src, cache_directory)
        results = {}
        src = Path(src)
        out = Path(out)
        files = [f for f in os.listdir(src) if f.endswith(".gpickle")]
        for graph_path in tqdm(
            files, desc=f"Solving all given instances using " + str(self)
        ):
            graph_path = graph_path.replace(".gpickle", "")
            if self.weighted:
                executable = (
                    self.kamis_path / "KaMIS" / "deploy" / "weighted_branch_reduce"
                )
            else:
                executable = self.kamis_path / "KaMIS" / "deploy" / "redumis"
            _preprocessed_graph = os.path.join(
                cache_directory,
                (
                    graph_path
                    + f"_{'weighted' if self.weighted else 'unweighted'}.graph"
                ),
            )
            results_filename = os.path.join(
                out,
                (
                    graph_path
                    + f"_{'weighted' if self.weighted else 'unweighted'}.result"
                ),
            )
            arguments = [
                _preprocessed_graph,  # input
                "--output",
                results_filename,  # output
                "--time_limit",
                str(self.time_limit),
            ]

            start_time = time.monotonic()
            result = subprocess.run(
                [executable] + arguments, shell=False, capture_output=True, text=True
            )
            lines = result.stdout.split("\n")
            solve_time = time.monotonic() - start_time

            results[graph_path] = {"total_time": solve_time}
            with open(results_filename, "r") as f:
                vertices = list(map(int, f.read().replace("\n", "")))
            is_vertices = np.flatnonzero(np.array(vertices))

            if self.weighted:
                discovery = re.compile("^(\d+(\.\d*)?) \[(\d+\.\d*)\]$")
                max_mwis_weight = 0
                mis_time = 0.0
                for line in lines:
                    match = discovery.match(line)
                    if match:
                        mwis_weight = float(match[1])
                        if mwis_weight > max_mwis_weight:
                            max_mwis_weight = mwis_weight
                            mis_time = float(match[3])

                if max_mwis_weight == 0:
                    # try another method
                    for line in lines:
                        if line.startswith("time"):
                            mis_time = line.split(" ")[1]
                        if line.startswith("MIS_weight"):
                            max_mwis_weight = line.split(" ")[1]

                if max_mwis_weight == 0:
                    results[graph_path]["mwis_found"] = False
                else:
                    results[graph_path]["mwis_found"] = True
                    results[graph_path]["mwis"] = is_vertices.tolist()
                    results[graph_path]["time_to_find_mwis"] = mis_time
                    results[graph_path]["mwis_vertices"] = is_vertices.shape[0]
                    results[graph_path]["mwis_weight"] = max_mwis_weight

            else:
                stdout = "\n".join(lines)
                discovery = re.compile(
                    "Best solution:\s+(\d+)\nTime:\s+(\d+\.\d*)\n", re.MULTILINE
                )
                time_found_in_stdout = False
                solution_time = 0.0
                for size, timestamp in discovery.findall(stdout):
                    if int(size) == is_vertices.shape[0]:
                        solution_time = float(timestamp)
                        time_found_in_stdout = True
                        break

                if not time_found_in_stdout:
                    # try another regex
                    discovery = re.compile(
                        "Best\n={42}\nSize:\s+\d+\nTime found:\s+(\d+\.\d*)",
                        re.MULTILINE,
                    )
                    m = discovery.search(stdout)
                    if m:
                        solution_time = float(m.group(1))
                        time_found_in_stdout = True

                if not time_found_in_stdout:
                    results[graph_path]["found_mis"] = False
                else:
                    results[graph_path]["found_mis"] = True
                    results[graph_path]["mis"] = is_vertices.tolist()
                    results[graph_path]["vertices"] = is_vertices.shape[0]
                    results[graph_path]["solution_time"] = solution_time

            with open(out / "results.json", "w", encoding="utf-8") as f:
                json.dump(results, f, ensure_ascii=False, sort_keys=True, indent=4)

    def recompile_kamis(self):
        if os.path.exists(self.kamis_path / "KaMIS/deploy/"):
            shutil.rmtree(self.kamis_path / "KaMIS/deploy/")
        if os.path.exists(self.kamis_path / "KaMIS/tmp_build/"):
            shutil.rmtree(self.kamis_path / "KaMIS/tmp_build/")
        shutil.copytree(
            self.kamis_path / "kamis-source/", self.kamis_path / "KaMIS/tmp_build/"
        )
        ori_dir = os.getcwd()
        os.chdir(self.kamis_path / "KaMIS/tmp_build/")
        os.system("bash cleanup.sh")
        os.system("bash compile_withcmake.sh")
        os.chdir(ori_dir)
        shutil.copytree(
            self.kamis_path / "KaMIS/tmp_build/deploy/",
            self.kamis_path / "KaMIS/deploy/",
        )
        shutil.rmtree(self.kamis_path / "KaMIS/tmp_build/")

    def __str__(self) -> str:
        return "kamis"



import os
import sys
import math
import numpy as np
from typing import Union
from pyvrp import Model
from pyvrp import read as read_vrp
from pyvrp import read_solution
from pyvrp.read import ROUND_FUNCS
from ml4co_kit.utils.type import to_numpy
from ml4co_kit.evaluate.cvrp.base import CVRPEvaluator
from ml4co_kit.evaluate.tsp.base import geographical


SUPPORT_NORM_TYPE = ["EUC_2D", "GEO"]
if sys.version_info.major == 3 and sys.version_info.minor == 8:
    CP38 = True
else:
    CP38 = False


class CVRPSolver:
    def __init__(
        self, 
        solver_type: str = None, 
        depots_scale: int = 1e6,
        points_scale: int = 1e6,
        demands_scale: int = 1e6,
        capacities_scale: int = 1e6,
    ):
        self.solver_type = solver_type
        self.depots_scale = depots_scale
        self.points_scale = points_scale
        self.demands_scale = demands_scale
        self.capacities_scale = capacities_scale
        self.depots = None
        self.ori_depots = None
        self.points = None
        self.ori_points = None
        self.demands = None
        self.capacities = None
        self.tours = None
        self.ref_tours = None
        self.nodes_num = None

    def check_depots_dim(self):
        if self.depots is None:
            return
        elif self.depots.ndim == 1:
            self.depots = np.expand_dims(self.depots, axis=0)
        if self.depots.ndim != 2:
            raise ValueError("The dimensions of ``depots`` cannot be larger than 2.")
    
    def check_ori_depots_dim(self):
        self.check_depots_dim()
        if self.ori_depots is None:
            return
        elif self.ori_depots.ndim == 1:
            self.ori_depots = np.expand_dims(self.ori_depots, axis=0)
        if self.ori_depots.ndim != 2:
            raise ValueError("The dimensions of ``ori_depots`` cannot be larger than 2.")
    
    def check_points_dim(self):
        if self.points is None:
            return
        elif self.points.ndim == 2:
            self.points = np.expand_dims(self.points, axis=0)
        if self.points.ndim != 3:
            raise ValueError("``points`` must be a 2D or 3D array.")
        self.nodes_num = self.points.shape[1]

    def check_ori_points_dim(self):
        self.check_points_dim()
        if self.ori_points is None:
            return
        elif self.ori_points.ndim == 2:
            self.ori_points = np.expand_dims(self.ori_points, axis=0)
        if self.ori_points.ndim != 3:
            raise ValueError("The ``ori_points`` must be 2D or 3D array.")
        
    def check_demands_dim(self):
        if self.demands is None:
            return
        elif self.demands.ndim == 1:
            self.demands = np.expand_dims(self.demands, axis=0)
        if self.demands.ndim != 2:
            raise ValueError("The dimensions of ``demands`` cannot be larger than 2.")
    
    def check_capacities_dim(self):
        if self.capacities is None:
            return
        if self.capacities.ndim != 1:
            raise ValueError("The ``capacities`` must be 1D array.")
    
    def check_tours_dim(self):
        if self.tours is None:
            return
        elif self.tours.ndim == 1:
            self.tours = np.expand_dims(self.tours, axis=0)
        if self.tours.ndim != 2:
            raise ValueError("The dimensions of ``tours`` cannot be larger than 2.")

    def check_ref_tours_dim(self):
        if self.ref_tours is None:
            return
        elif self.ref_tours.ndim == 1:
            self.ref_tours = np.expand_dims(self.ref_tours, axis=0)
        if self.ref_tours.ndim != 2:
            raise ValueError(
                "The dimensions of the ``ref_tours`` cannot be larger than 2."
            )

    def check_depots_not_none(self):
        if self.depots is None:
            message = (
                "``depots`` cannot be None! You can load the points using the methods"
                "``from_data``, ``from_txt``, or ``from_vrp``."
            )
            raise ValueError(message)
         
    def check_points_not_none(self):
        if self.points is None:
            message = (
                "``points`` cannot be None! You can load the points using the methods"
                "``from_data``, ``from_txt``, or ``from_vrp``."
            )
            raise ValueError(message)

    def check_demands_not_none(self):
        if self.demands is None:
            message = (
                "``demands`` cannot be None! You can load the points using the methods"
                "``from_data``, ``from_txt``, or ``from_vrp``."
            )
            raise ValueError(message)
    
    def check_capacities_not_none(self):
        if self.demands is None:
            message = (
                "``capacities`` cannot be None! You can load the points using the methods"
                "``from_data``, ``from_txt``, or ``from_vrp``."
            )
            raise ValueError(message)
    
    def check_tours_not_none(self):
        if self.tours is None:
            message = (
                "``tours`` cannot be None! You can use solvers based on ``CVRPSolver`` "
                "like ``CVRPPyVRPSolver`` or the method ``read_tours`` to get the tours."
            )
            raise ValueError(message)
        
    def check_ref_tours_not_none(self):
        if self.ref_tours is None:
            message = (
                "``ref_tours`` cannot be None! You can use solvers based on ``CVRPSolver`` "
                "like ``CVRPPyVRPSolver`` to obtain the tours and set them as the ``ref_tours``. "
                "Or you can use the methods ``read_ref_tours``, ``from_txt``, "
                "or ``read_ref_tours_from_sol`` to load them."
            )
            raise ValueError(message)
    
    def get_round_func(self, round_func: str):
        if (key := str(round_func)) in ROUND_FUNCS:
            round_func = ROUND_FUNCS[key]
        if not callable(round_func):
            raise TypeError(
                f"round_func = {round_func} is not understood. Can be a function,"
                f" or one of {ROUND_FUNCS.keys()}."
            )
        return round_func
    
    def set_norm(self, norm: str):
        if norm is None:
            return
        if norm not in SUPPORT_NORM_TYPE:
            message = (
                f"The norm type ({norm}) is not a valid type, "
                f"only {SUPPORT_NORM_TYPE} are supported."
            )
            raise ValueError(message)
        if norm == "GEO" and self.scale != 1:
            message = "The scale must be 1 for ``GEO`` norm type."
            raise ValueError(message)
        self.norm = norm

    def normalize_points_depots(self):
        for idx in range(self.points.shape[0]):
            cur_points = self.points[idx]
            cur_depots = self.depots[idx]
            max_value = np.max(cur_points)
            min_value = np.min(cur_points)
            cur_points = (cur_points - min_value) / (max_value - min_value)
            cur_depots = (cur_depots - min_value) / (max_value - min_value)
            self.points[idx] = cur_points
            self.depots[idx] = cur_depots
    
    def check_demands_meet(
        self,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[list, np.ndarray] = None,
        tours: Union[np.ndarray, list] = None,
    ):
        self.from_data(demands=demands, capacities=capacities)
        self.read_tours(tours)
        tours_shape = self.tours.shape
        for idx in range(tours_shape[0]):
            cur_demand = self.demands[idx]
            cur_capacity = self.capacities[idx]
            cur_tour = self.tours[idx]
            split_tours = np.split(cur_tour, np.where(cur_tour == 0)[0])[1: -1]
            for split_idx in range(len(split_tours)):
                split_tour = split_tours[split_idx][1:]
                split_demand_need = np.sum(cur_demand[split_tour.astype(int) - 1], dtype=np.float32)
                if split_demand_need > cur_capacity + 1e-5:
                    message = (
                        f"Capacity constraint not met in tour {idx}. "
                        f"The split tour is ``{split_tour}`` with the demand of {split_demand_need}."
                        f"However, the maximum capacity of the vehicle is {cur_capacity}."
                    )
                    raise ValueError(message)

    def get_distance(self, x1: float, x2: float, norm: str = None):
        self.set_norm(norm)
        if self.norm == "EUC_2D":
            return math.sqrt((x1[0] - x2[0]) ** 2 + (x1[1] - x2[1]) ** 2)
        elif self.norm == "GEO":
            return geographical(x1, x2)
                    
    def from_data(
        self,
        depots: Union[list, np.ndarray] = None,
        points: Union[list, np.ndarray] = None,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[int, float, np.ndarray] = None,
        norm: str = "EUC_2D",
        normalize: bool = False
    ):
        # norm
        self.set_norm(norm)
        
        # depots
        if depots is not None:
            depots = to_numpy(depots)
            self.ori_depots = depots
            self.depots = depots.astype(np.float32)
            self.check_ori_depots_dim()
        
        # points
        if points is not None:
            points = to_numpy(points)
            self.ori_points = points
            self.points = points.astype(np.float32)
            self.check_ori_points_dim()
        
        # demands
        if demands is not None:
            demands = to_numpy(demands)
            self.demands = demands.astype(np.float32)
            self.check_demands_dim()
        
        # capacities
        if capacities is not None:
            if type(capacities) == float or type(capacities) == int:
                capacities = np.array([capacities])
            self.capacities = capacities.astype(np.float32)
            self.capacities = capacities
            self.check_capacities_dim()

        # normalize
        if normalize:
            self.normalize_points_depots()
    
    def from_txt(
        self,
        file_path: str,
        norm: str = "EUC_2D",
        normalize: bool = False,
        load_ref_tours: str = True,
        return_list: bool = False,
    ):
        self.set_norm(norm)

        # check the file format
        if not file_path.endswith(".txt"):
            raise ValueError("Invalid file format. Expected a ``.txt`` file.")

        # read the data form .txt
        with open(file_path, "r") as file:
            # record to lists
            depot_list = list()
            points_list = list()
            demands_list = list()
            capacity_list = list()
            ref_tours = list()
            
            # read by lines
            for line in file:
                # line to strings
                line = line.strip()
                split_line_0 = line.split("depots ")[1]
                split_line_1 = split_line_0.split(" points ")
                depot = split_line_1[0]
                split_line_2 = split_line_1[1].split(" demands ")
                points = split_line_2[0]
                split_line_3 = split_line_2[1].split(" capacity ")
                demands = split_line_3[0]
                split_line_4 = split_line_3[1].split(" output ")
                capacity = split_line_4[0]
                ref_tour = split_line_4[1]
                
                # strings to array
                depot = depot.split(" ")
                depot = np.array([float(depot[0]), float(depot[1])])
                points = points.split(" ")
                points = np.array(
                    [
                        [float(points[i]), float(points[i + 1])]
                        for i in range(0, len(points), 2)
                    ]
                )
                demands = demands.split(" ")
                demands = np.array([
                    float(demands[i]) for i in range(len(demands))
                ])
                capacity = float(capacity)
                ref_tour = ref_tour.split(" ")
                ref_tour = [int(t) for t in ref_tour]
                
                # add to the list
                depot_list.append(depot)
                points_list.append(points)
                demands_list.append(demands)
                capacity_list.append(capacity)
                ref_tours.append(ref_tour)

        if return_list:
            return depot_list, points_list, demands_list, capacity_list, ref_tours
        
        if load_ref_tours:
            self.read_ref_tours(ref_tours)
        
        depots = np.array(depot_list)
        self.ori_depots = depots
        self.depots = depots.astype(np.float32)
        nodes_coords = np.array(points_list)
        self.ori_points = nodes_coords
        self.points = nodes_coords.astype(np.float32)
        self.demands = np.array(demands_list).astype(np.float32)
        self.capacities = np.array(capacity_list).astype(np.float32)
        self.check_ori_depots_dim()
        self.check_ori_points_dim()
        self.check_ref_tours_dim()
        self.check_demands_dim()
        self.check_capacities_dim()
        if normalize:
            self.normalize_points_depots()
    
    def from_vrp(
        self, 
        file_path: str, 
        round_func: str = "none",
        norm: str = "EUC_2D",
        normalize: bool = False
    ):
        instance = read_vrp(where=file_path, round_func=round_func)
        vrp_model = Model.from_data(instance)
        _depots = vrp_model._depots[0]
        _clients = vrp_model._clients
        _vehicle_types = vrp_model._vehicle_types[0]
        depots = np.array([_depots.x, _depots.y])
        points_list = list()
        demands_list = list()
        for client in _clients:
            points_list.append([client.x, client.y])
            demands_list.append(client.demand if CP38 else client.delivery)
        capacity = _vehicle_types.capacity
        self.from_data(
            depots=depots,
            points=points_list,
            demands=demands_list,
            capacities=capacity,
            norm=norm,
            normalize=normalize
        )        
     
    def read_tours(self, tours: Union[list, np.ndarray]):
        if tours is None:
            return
        if type(tours) == list:
            # 1D tours
            if type(tours[0]) != list and type(tours[0]) != np.ndarray:
                tours = np.array(tours)
            # 2D tours
            else:
                lengths = [len(tour) for tour in tours]
                max_length = max(lengths)
                len_tours = len(tours)
                np_tours = np.zeros(shape=(len_tours, max_length)) - 1
                for idx in range(len_tours):
                    tour = tours[idx]
                    len_tour = len(tour)
                    np_tours[idx][:len_tour] = tour
                tours = np_tours
        self.tours = tours.astype(np.int32)
        self.check_tours_dim()
    
    def read_ref_tours(self, ref_tours: Union[list, np.ndarray]):
        if ref_tours is None:
            return
        if type(ref_tours) == list:
            # 1D tours
            if type(ref_tours[0]) != list and type(ref_tours[0]) != np.ndarray:
                ref_tours = np.array(ref_tours)
            # 2D tours
            else:
                lengths = [len(ref_tours) for ref_tours in ref_tours]
                max_length = max(lengths)
                len_ref_tours = len(ref_tours)
                np_ref_tours = np.zeros(shape=(len_ref_tours, max_length)) - 1
                for idx in range(len_ref_tours):
                    ref_tour = ref_tours[idx]
                    len_ref_tour = len(ref_tour)
                    np_ref_tours[idx][:len_ref_tour] = ref_tour
                ref_tours = np_ref_tours
        self.ref_tours = ref_tours.astype(np.int32)
        self.check_ref_tours_dim()

    def read_ref_tours_from_sol(self, file_path: str):
        if not file_path.endswith(".sol"):
            raise ValueError("Invalid file format. Expected a ``.sol`` file.")
        
        # check the .sol type
        route_flag = None
        with open(file_path, "r") as file:
            first_line = file.readline()
            if "Route" in first_line:
                # Like this
                # Route #1: 15 17 9 3 16 29
                # Route #2: 12 5 26 7 8 13 32 2
                route_flag = True
            else:
                # Like this
                # 36395
                # 37
                # 1893900
                # 1133620
                # 0 1 1 1144 12  14 0 217 236 105 2 169 8 311 434 362 187 136 59 0
                # 0 1 1 1182 12  14 0 3 370 133 425 349 223 299 386 267 410 411 348 0
                route_flag = False
        
        # read the data form .sol
        if route_flag == True:
            solution = read_solution(where=file_path)
            tour = [0]
            for _solution in solution:
                tour = tour + _solution + [0]
        elif route_flag == False:
            with open(file_path, "r") as file:
                line_idx = 0
                tour = [0]
                for line in file:
                    line_idx += 1
                    if line_idx < 5:
                        continue
                    split_line = line.split(" ")[7:-1]
                    for node in split_line:
                        tour.append(int(node))
                    tour.append(0)
        else:
            raise ValueError(f"Unable to read route information from {file_path}.")
        
        # read ref tours
        self.read_ref_tours(tour)
        
    def to_txt(
        self,
        filename: str = "example.txt",
        original: bool = True,
        depots: Union[list, np.ndarray] = None,
        points: Union[list, np.ndarray] = None,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[int, float, np.ndarray] = None,
        tours: Union[np.ndarray, list] = None,
        norm: str = None,
        normalize: bool = False,
    ):
        # read and check
        self.from_data(depots, points, demands, capacities, norm, normalize)
        self.read_tours(tours)
        if self.tours is None:
            raise ValueError(
                "``tours`` cannot be None, please use method 'solve' to get solutions."
            )
        self.check_depots_not_none()
        self.check_points_not_none()
        self.check_demands_not_none()
        self.check_capacities_not_none()
        self.check_tours_not_none()
        
        # variables
        _depots = self.ori_depots if original else self.depots
        _points = self.ori_points if original else self.points
        _demands = self.demands
        _capacities = self.capacities
        _tours = self.tours

        # write
        with open(filename, "w") as f:
            # write to txt
            for idx in range(len(_tours)):
                tour = _tours[idx]
                tour = np.split(tour, np.where(tour == -1)[0])[0]
                depot = _depots[idx]
                points = _points[idx]
                demands = _demands[idx]
                capicity = _capacities[idx]
                f.write("depots " + str(depot[0]) + str(" ") + str(depot[1]))
                f.write(" points" + str(" "))
                f.write(
                    " ".join(
                        str(x) + str(" ") + str(y)
                        for x, y in points
                    )
                )
                f.write(" demands " + str(" ").join(str(demand) for demand in demands))
                f.write(" capacity " + str(capicity))
                f.write(str(" output "))
                f.write(str(" ").join(str(node_idx) for node_idx in tour.tolist()))
                f.write("\n")
            f.close()

    def to_vrp(
        self,
        save_dir: str,
        filename: str,
        original: bool = True,
        depots: Union[list, np.ndarray] = None,
        points: Union[list, np.ndarray] = None,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[int, float, np.ndarray] = None,
        norm: str = None,
        normalize: bool = False,
        dtype: str = "int",
        round_func: str = "round"
    ):
        # prepare
        self.from_data(depots, points, demands, capacities, norm, normalize)
        if filename.endswith(".vrp"):
            filename = filename.replace(".vrp", "")
        self.check_depots_not_none()
        self.check_points_not_none()
        self.check_demands_not_none()
        self.check_capacities_not_none()
        
        # variables
        depots = self.ori_depots if original else self.ori_depots
        points = self.ori_points if original else self.points
        demands = self.demands
        capacities = self.capacities
        if dtype == "int":
            self.round_func = self.get_round_func(round_func)
            depots = self.round_func(depots)
            points = self.round_func(points)
            demands = self.round_func(demands)
            capacities = self.round_func(capacities)
            
        # makedirs
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        # write
        for idx in range(points.shape[0]):
            save_path = os.path.join(save_dir, filename + f"-{idx}.vrp")
            with open(save_path, "w") as f:
                f.write(f"NAME : {save_path}\n")
                f.write("TYPE : CVRP\n")
                f.write(f"DIMENSION : {self.nodes_num + 1}\n")
                f.write(f"EDGE_WEIGHT_TYPE : {self.norm}\n")
                f.write(f"CAPACITY : {self.capacities[idx]}\n")
                f.write("NODE_COORD_SECTION\n")
                x, y = depots[idx]
                f.write(f"1 {x} {y}\n")
                for i in range(self.nodes_num):
                    x, y = points[idx][i]
                    f.write(f"{i+2} {x} {y}\n")
                f.write("DEMAND_SECTION \n")
                f.write(f"1 0\n")
                for i in range(self.nodes_num):
                    f.write(f"{i+2} {demands[idx][i]}\n")
                f.write("DEPOT_SECTION \n")
                f.write("	1\n")
                f.write("	-1\n")
                f.write("EOF\n")
    
    def to_sol(
        self,
        save_dir: str,
        filename: str,
        tours: Union[np.ndarray, list] = None,
        dtype: str = "int",
        round_func: str = "round"
    ):
        # read and check
        self.read_tours(tours)
        if filename.endswith(".sol"):
            filename = filename.replace(".sol", "")
        self.check_tours_not_none()
        tours = self.tours
        
        # makedirs
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        # write
        for idx in range(tours.shape[0]):
            save_path = os.path.join(save_dir, filename + f"-{idx}.sol")
            tour = tours[idx]
            split_tours = np.split(tour, np.where(tour == 0)[0])[1: -1]
            with open(save_path, "w") as f:
                for i in range(len(split_tours)):
                    part_tour = split_tours[i][1:]
                    f.write(f"Route #{i+1}: ")
                    f.write(f" ".join(str(int(node)) for node in part_tour))
                    f.write("\n")
                cost = self.evaluate(
                    calculate_gap=False, 
                    dtype=dtype, 
                    round_func=round_func
                )
                f.write(f"Cost {cost}\n")
    
    def modify_tour(self, tour: np.ndarray):
        if not np.isin(-1, tour):
            return tour
        return tour[: np.where(tour == -1)[0][0]]
    
    def evaluate(
        self,
        original: bool = True,
        depots: Union[list, np.ndarray] = None,
        points: Union[list, np.ndarray] = None,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[list, np.ndarray] = None,
        norm: str = None,
        normalize: bool = False,
        tours: Union[np.ndarray, list] = None,
        ref_tours: Union[np.ndarray, list] = None,
        calculate_gap: bool = False,
        check_demands: bool = True,
        dtype: str = "float",
        round_func: str = "none"
    ):
        # read and check
        self.from_data(depots, points, demands, capacities, norm, normalize)
        self.read_tours(tours)
        self.read_ref_tours(ref_tours)
        self.check_points_not_none()
        self.check_tours_not_none()
        if check_demands:
            self.check_demands_meet()
        if calculate_gap:
            self.check_ref_tours_not_none()
        depots = self.ori_depots if original else self.depots
        points = self.ori_points if original else self.points
        tours = self.tours
        ref_tours = self.ref_tours

        # prepare for evaluate
        tours_cost_list = list()
        samples = points.shape[0]
        if calculate_gap:
            ref_tours_cost_list = list()
            gap_list = list()
            
        # evaluate
        for idx in range(samples):
            evaluator = CVRPEvaluator(depots[idx], points[idx], self.norm)
            solved_tour = tours[idx]
            solved_cost = evaluator.evaluate(self.modify_tour(solved_tour), dtype, round_func)
            tours_cost_list.append(solved_cost)
            if calculate_gap:
                ref_cost = evaluator.evaluate(self.modify_tour(ref_tours[idx]), dtype, round_func)
                ref_tours_cost_list.append(ref_cost)
                gap = (solved_cost - ref_cost) / ref_cost * 100
                gap_list.append(gap)

        # calculate average cost/gap & std
        tours_costs = np.array(tours_cost_list)
        if calculate_gap:
            ref_costs = np.array(ref_tours_cost_list)
            gaps = np.array(gap_list)
        costs_avg = np.average(tours_costs)
        if dtype == "int":
            costs_avg = int(costs_avg)
        if calculate_gap:
            ref_costs_avg = np.average(ref_costs)
            gap_avg = np.sum(gaps) / samples
            gap_std = np.std(gaps)
            return costs_avg, ref_costs_avg, gap_avg, gap_std
        else:
            return costs_avg
        
    def solve(
        self,
        depots: Union[list, np.ndarray] = None,
        points: Union[list, np.ndarray] = None,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[list, np.ndarray] = None,
        norm: str = "EUC_2D",
        normalize: bool = False,
        num_threads: int = 1,
        show_time: bool = False,
        **kwargs,
    ) -> np.ndarray:
        raise NotImplementedError(
            "The ``solve`` function is required to implemented in subclasses."
        )
    


import time
import numpy as np
from tqdm import tqdm
from typing import Union, Any
from pytorch_lightning.utilities import rank_zero_info
from ml4co_kit.solver import TSPSolver
from ml4co_kit.learning.model import BaseModel, ModelConfigurator
from ml4co_kit.learning.search import SearchConfigurator
from model import TSPGNN
from search import tsp_greedy, tsp_2opt


class TSPNARSolver(TSPSolver):
    """
    This is an TSP solver example to easily develop a solver class based on the learning models
    by inheriting from the class ``solver.TSPSolver``.
    """

    def __init__(self):
        super(TSPNARSolver, self).__init__()

    def solve(
        self,
        batch_size: int = 16,
        encoder: Union[BaseModel, str] = "gnn",
        encoder_kwargs: dict = {},
        decoding_type: Union[Any, str] = "greedy",
        decoding_kwargs: dict = {},
        local_search_type: str = "2opt",
        ls_kwargs: dict = {},
        pretrained: bool = True,
        device="cpu",
    ):
        self.decoding_type = decoding_type
        self.ls_type = local_search_type
        self.model_config = ModelConfigurator(model_class={("tsp", "gnn"): TSPGNN})
        self.search_config = SearchConfigurator(
            decoding_class={("tsp", "greedy"): tsp_greedy},
            local_search_class={("tsp", "2opt"): tsp_2opt},
        )

        # encoder & gain heatmap
        if type(encoder) == str:
            encoder_kwargs.update(
                {
                    "mode": "solve",
                    "nodes_num": self.nodes_num,
                }
            )
            self.encoder = self.model_config.get_model(task="tsp", model_name=encoder)(
                **encoder_kwargs
            )
        else:
            self.encoder = encoder
        rank_zero_info(f"Begin encoding, Using {self.encoder}")
        if pretrained:
            rank_zero_info(f"Loading Weights from Pretrained CheckPoint")
            pretrained_path = (
                encoder_kwargs["pretrained_path"]
                if "pretrained_path" in encoder_kwargs.keys()
                else None
            )
            self.encoder.load_weights(pretrained_path)
        self.encoder.to(device)

        # model inference
        solve_begin_time = time.time()
        edge_index = None
        points = self.points
        heatmap = self.encoder.solve(points, edge_index, batch_size, device)
        solve_end_time = time.time()
        solve_time = solve_end_time - solve_begin_time
        rank_zero_info(f"Model Inference, Using {solve_time}")

        # decoding
        if type(decoding_type) == str:
            self.decoding_func = self.search_config.get_decoding_func(
                task="tsp", name=decoding_type
            )
        else:
            self.decoding_func = decoding_type
        rank_zero_info(f"Decoding, Using {self.decoding_func.__name__}")
        decoded_tours = list()
        for idx in tqdm(range(self.points.shape[0]), desc="Decoding"):
            adj_mat = np.expand_dims(heatmap[idx], axis=0)
            tour = self.decoding_func(
                adj_mat=adj_mat, 
                np_points=self.points[idx], 
                edge_index_np=None,
                **decoding_kwargs
            )
            decoded_tours.append(tour[0])
        decoded_tours = np.array(decoded_tours)

        # local_search
        ls_tours = None
        self.local_search_func = self.search_config.get_local_search_func(
            task="tsp", name=local_search_type
        )
        if self.local_search_func is not None:
            rank_zero_info(f"Local Search, Using {self.local_search_func.__name__}")
            ls_tours = list()
            for idx in tqdm(range(self.points.shape[0]), desc="Local Search"):
                adj_mat = heatmap[idx]
                tour = self.local_search_func(
                    np_points=self.points[idx],
                    tours=decoded_tours[idx],
                    adj_mat=adj_mat,
                    device=device,
                    **ls_kwargs,
                )
                ls_tours.append(tour)
            ls_tours = np.array(ls_tours)

        tours = decoded_tours if ls_tours is None else ls_tours
        self.tours = tours
        return tours

    def __repr__(self):
        message = f"encoder={self.encoder}, decoding_type={self.decoding_type}, ls_type={self.ls_type}"
        return f"{self.__class__.__name__}({message})"



import os
from ml4co_kit.utils.file_utils import _get_md5


class ML4TSPDataset:
    def __init__(self):
        self.supported = {
            "tsp50_uniform_lkh_5k_1.28m.txt": "d049c9e86d7b26d2e6d25d896e0545fd",
            "tsp100_uniform_lkh_5k_1.28m.txt": "787bb9c5481cfa7835d640b702f9fe82",
            "tsp500_uinform_concorde_80k.txt": "da01970bdb52aa7874f92222f09b0296", 
            "tsp500_uniform_lkh_50k_128k.txt": "833ef74de9d982b8f7420b001dd628fa",
            "tsp1000_uniform_lkh_100k_64k.txt": "96ff21d0fcf75a7cc9dd2e4fbf8f8789",
            "tsp10000_uniform_concorde_large_6.4k.txt": "3e0e379ca6b49f30e2bad935176cea02"
        }
        self.dir = "dataset/ml4tsp"
        if not os.path.exists(self.dir):
            os.makedirs(self.dir)

    def download(
        self,
        filename: str,
        hf_token: str
    ):
        # check necessary package
        try:
            from huggingface_hub import HfApi
        except:
            raise ModuleNotFoundError("huggingface_hub need be installed!")
        
         # use huggingface token to download
        hf_api = HfApi(token=hf_token)
        hf_api.hf_hub_download(
            repo_id="ML4CO/ML4TSPDataset",
            repo_type="dataset",
            filename=filename,
            cache_dir="hf_cache",
            local_dir=self.dir,
            local_dir_use_symlinks=False
        )
        
        # check the downloaded file (md5)
        self.save_path = os.path.join(self.dir, filename)
        md5 = _get_md5(self.save_path)
        if md5 != self.supported[filename]:
            os.remove(self.save_path)
            raise ValueError("Warning: MD5 check failed for the downloaded content")
            


import os
import time
import uuid
import pathlib
import tsplib95
import numpy as np
from tqdm import tqdm
from subprocess import check_call
from multiprocessing import Pool
from typing import Union
from .base import CVRPSolver


class CVRPLKHSolver(CVRPSolver):
    def __init__(
        self,
        depots_scale: int = 1e6,
        points_scale: int = 1e6,
        demands_scale: int = 1e6,
        capacities_scale: int = 1e6,
        lkh_max_trials: int = 1000,
        lkh_path: pathlib.Path = "LKH",
        lkh_runs: int = 10,
        lkh_seed: int = 1234,
    ):
        """
        TSPLKHSolver
        Args:
            lkh_max_trials (int, optional): The maximum number of trials for
                the LKH solver. Defaults to 1000.
            lkh_path (pathlib.Path, optional): The path to the LKH solver.
                Defaults to "LKH".
            scale (int, optional): The scale factor for coordinates in the
                LKH solver. Defaults to 1e6.
            lkh_runs (int, optional): The number of runs for the LKH solver.
                Defaults to 10.
        """
        super(CVRPLKHSolver, self).__init__(
            solver_type="lkh", 
            depots_scale = depots_scale,
            points_scale = points_scale,
            demands_scale = demands_scale,
            capacities_scale = capacities_scale,
        )
        self.lkh_max_trials = lkh_max_trials
        self.lkh_path = lkh_path
        self.lkh_runs = lkh_runs
        self.lkh_seed = lkh_seed

    def write_parameter_file(
        self,
        save_path: str,
        vrp_file_path: str,
        tour_path: str
    ):
        with open(save_path, "w") as f:
            f.write(f"PROBLEM_FILE = {vrp_file_path}\n")
            f.write(f"MAX_TRIALS = {self.lkh_max_trials}\n")
            f.write(f"SPECIAL\nRUNS = {self.lkh_runs}\n")
            f.write(f"SEED = {self.lkh_seed}\n")
            f.write(f"TOUR_FILE = {tour_path}\n")
    
    def read_lkh_solution(self, tour_path: str) -> list:
        tour = tsplib95.load(tour_path).tours[0]
        np_tour = np.array(tour) - 1
        over_index = np.where(np_tour > self.nodes_num)[0]
        np_tour[over_index] = 0
        tour = np_tour.tolist()
        tour: list
        tour.append(0)
        return tour
        
    def _solve(
        self, 
        depot_coord: np.ndarray, 
        nodes_coord: np.ndarray,
        demands: np.ndarray,
        capacity: float
    ) -> list:
        # scale
        depot_coord = (depot_coord * self.depots_scale).astype(np.int64)
        nodes_coord = (nodes_coord * self.points_scale).astype(np.int64)
        demands = (demands * self.demands_scale).astype(np.int64)
        capacity = int(capacity * self.capacities_scale)
        
        # Intermediate files
        tmp_name = uuid.uuid4().hex[:9]
        para_save_path = f"{tmp_name}.para"
        vrp_save_path = f"{tmp_name}.vrp"
        real_vrp_save_path = f"{tmp_name}-0.vrp"
        tour_save_path = f"{tmp_name}.tour"
        log_save_path = f"{tmp_name}.log"
        
        # prepare for solve
        self.tmp_solver.from_data(
            depots=depot_coord,
            points=nodes_coord,
            demands=demands,
            capacities=capacity
        )
        self.tmp_solver.to_vrp(
            save_dir="./", 
            filename=vrp_save_path
        )
        self.write_parameter_file(
            save_path=para_save_path,
            vrp_file_path=real_vrp_save_path,
            tour_path=tour_save_path
        )
        
        # solve
        with open(log_save_path, "w") as f:
            check_call([self.lkh_path, para_save_path], stdout=f)
            
        # read solution
        tour = self.read_lkh_solution(tour_save_path)
        
        # delete files
        files_path = [
            para_save_path, real_vrp_save_path,
            tour_save_path, log_save_path
        ]
        for file_path in files_path:
           if os.path.exists(file_path):
               os.remove(file_path)
        
        # return
        return tour
            
    def solve(
        self,
        depots: Union[list, np.ndarray] = None,
        points: Union[list, np.ndarray] = None,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[list, np.ndarray] = None,
        norm: str = "EUC_2D",
        normalize: bool = False,
        num_threads: int = 1,
        show_time: bool = False,
    ) -> np.ndarray:
        # prepare
        self.from_data(depots, points, demands, capacities, norm, normalize)
        self.tmp_solver = CVRPSolver()
        start_time = time.time()

        # solve
        tours = list()
        p_shape = self.points.shape
        num_points = p_shape[0]
        if num_threads == 1:
            if show_time:
                for idx in tqdm(range(num_points), desc="Solving CVRP Using LKH"):
                    tours.append(self._solve(
                        depot_coord=self.depots[idx],
                        nodes_coord=self.points[idx],
                        demands=self.demands[idx],
                        capacity=self.capacities[idx]
                    ))
            else:
                for idx in range(num_points):
                    tours.append(self._solve(
                        depot_coord=self.depots[idx],
                        nodes_coord=self.points[idx],
                        demands=self.demands[idx],
                        capacity=self.capacities[idx]
                    ))
        else:
            num_tqdm = num_points // num_threads
            batch_depots = self.depots.reshape(num_tqdm, num_threads, -1)
            batch_demands = self.demands.reshape(num_tqdm, num_threads, -1)
            batch_capacities = self.capacities.reshape(num_tqdm, num_threads)
            batch_points = self.points.reshape(-1, num_threads, p_shape[-2], p_shape[-1])
            if show_time:
                for idx in tqdm(
                    range(num_points // num_threads), desc="Solving CVRP Using LKH"
                ):
                    with Pool(num_threads) as p1:
                        cur_tours = p1.starmap(
                            self._solve,
                            [  (batch_depots[idx][inner_idx], 
                                batch_points[idx][inner_idx], 
                                batch_demands[idx][inner_idx], 
                                batch_capacities[idx][inner_idx]) 
                                for inner_idx in range(num_threads)
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)
            else:
                for idx in range(num_points // num_threads):
                    with Pool(num_threads) as p1:
                        cur_tours = p1.starmap(
                            self._solve,
                            [  (batch_depots[idx][inner_idx], 
                                batch_points[idx][inner_idx], 
                                batch_demands[idx][inner_idx], 
                                batch_capacities[idx][inner_idx]) 
                                for inner_idx in range(num_threads)
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)

        # format
        self.read_tours(tours)
        end_time = time.time()
        if show_time:
            print(f"Use Time: {end_time - start_time}")
        return self.tours    



import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from typing import Union


def edges_to_node_pairs(edge_target: np.ndarray):
    """Helper function to convert edge matrix into pairs of adjacent nodes."""
    pairs = []
    for r in range(len(edge_target)):
        for c in range(len(edge_target)):
            if edge_target[r][c] == 1:
                pairs.append((r, c))
    return pairs


def draw_tsp_problem(
    save_path: str,
    points: Union[list, np.ndarray],
    edge_values: np.ndarray = None,
    figsize: tuple = (5, 5),
    node_color: str = "darkblue",
    edge_color: str = "darkblue",
    node_size: int = 50,
):
    # check
    if "." not in save_path:
        save_path += ".png"
    if type(points) == list:
        points = np.array(points)
    if points.ndim == 3 and points.shape[0] == 1:
        points = points[0]
    if points.ndim != 2:
        raise ValueError("the dim of the points must 2.")

    # edge_values
    if edge_values is None:
        edge_values = (
            np.sum(
                (np.expand_dims(points, 1) - np.expand_dims(points, 0)) ** 2, axis=-1
            )
            ** 0.5
        )

    # edge_target
    nodes_num = points.shape[0]
    edge_target = np.zeros((nodes_num, nodes_num))
    target_pairs = edges_to_node_pairs(edge_target)
    graph = nx.from_numpy_array(edge_values)
    pos = dict(zip(range(len(points)), points.tolist()))

    # plt
    figure = plt.figure(figsize=figsize)
    figure.add_subplot(111)
    nx.draw_networkx_nodes(G=graph, pos=pos, node_color=node_color, node_size=node_size)
    nx.draw_networkx_edges(
        G=graph, pos=pos, edgelist=target_pairs, alpha=1, width=1, edge_color=edge_color
    )

    # save
    plt.savefig(save_path)


def draw_tsp_solution(
    save_path: str,
    points: Union[list, np.ndarray],
    tours: Union[list, np.ndarray],
    edge_values: np.ndarray = None,
    figsize: tuple = (5, 5),
    node_color: str = "darkblue",
    edge_color: str = "darkblue",
    node_size: int = 50,
):
    # check
    if "." not in save_path:
        save_path += ".png"
    if type(points) == list:
        points = np.array(points)
    if type(tours) == list:
        tours = np.array(tours)
    if points.ndim == 3 and points.shape[0] == 1:
        points = points[0]
    if tours.ndim == 2 and tours.shape[0] == 1:
        tours = tours[0]
    if points.ndim != 2:
        raise ValueError("the dim of the points must 2.")
    if tours.ndim != 1:
        raise ValueError("the dim of the tours must 1.")

    # edge_values
    if edge_values is None:
        edge_values = (
            np.sum(
                (np.expand_dims(points, 1) - np.expand_dims(points, 0)) ** 2, axis=-1
            )
            ** 0.5
        )

    # edge_target
    nodes_num = points.shape[0]
    edge_target = np.zeros((nodes_num, nodes_num))
    for i in range(len(tours) - 1):
        edge_target[tours[i], tours[i + 1]] = 1
    target_pairs = edges_to_node_pairs(edge_target)
    graph = nx.from_numpy_array(edge_values)
    pos = dict(zip(range(len(points)), points.tolist()))

    # plt
    figure = plt.figure(figsize=figsize)
    figure.add_subplot(111)
    nx.draw_networkx_nodes(G=graph, pos=pos, node_color=node_color, node_size=node_size)
    nx.draw_networkx_edges(
        G=graph, pos=pos, edgelist=target_pairs, alpha=1, width=1, edge_color=edge_color
    )

    # save
    plt.savefig(save_path)



import numpy as np
from typing import Union


def to_numpy(x: Union[np.ndarray, list]):
    if type(x) == list:
        return np.array(x)
    elif type(x) == np.ndarray:
        return x
    else:
        raise NotImplementedError()


import numpy as np
from typing import Union
from ml4co_kit.evaluate.tsp.base import TSPEvaluator


class CVRPEvaluator(TSPEvaluator):
    def __init__(
        self,
        depots: Union[list, np.ndarray],
        points: Union[list, np.ndarray], 
        norm: str = "EUC_2D"
    ):
        # depots
        if type(depots) == list:
            depots = np.array(depots)
        if depots.ndim == 2 and depots.shape[0] == 1:
            depots = depots[0]
        if depots.ndim != 1:
            raise ValueError("depots must be 1D array.")
        
        # points
        if type(points) == list:
            points = np.array(points)
        if points.ndim == 3 and points.shape[0] == 1:
            points = points[0]
        if points.ndim != 2:
            raise ValueError("points must be 2D array.")
        
        points_shape = points.shape
        coords = np.zeros(shape=(points_shape[0] + 1, points_shape[1]))
        coords[0] = depots
        coords[1:] = points
        self.points = coords
        self.set_norm(norm)


import sys
import time
import numpy as np
from tqdm import tqdm
from multiprocessing import Pool
from typing import Union
from pyvrp import Model
from pyvrp.stop import MaxRuntime
from .base import CVRPSolver


if sys.version_info.major == 3 and sys.version_info.minor == 8:
    CP38 = True
else:
    CP38 = False


class CVRPPyVRPSolver(CVRPSolver):
    def __init__(
        self,
        depots_scale: int = 1e6,
        points_scale: int = 1e6,
        demands_scale: int = 1e6,
        capacities_scale: int = 1e6,
        time_limit: float = 1.0,
    ):
        super(CVRPPyVRPSolver, self).__init__(
            solver_type="pyvrp", 
            depots_scale = depots_scale,
            points_scale = points_scale,
            demands_scale = demands_scale,
            capacities_scale = capacities_scale,
        )
        self.time_limit = time_limit

    def _solve(
        self, 
        depot_coord: np.ndarray, 
        nodes_coord: np.ndarray,
        demands: np.ndarray,
        capacity: float
    ) -> list:
        # scale
        depot_coord = (depot_coord * self.depots_scale).astype(np.int64)
        nodes_coord = (nodes_coord * self.points_scale).astype(np.int64)
        demands = (demands * self.demands_scale).astype(np.int64)
        capacity = int(capacity * self.capacities_scale)
        
        # solve
        cvrp_model = Model()
        depot = cvrp_model.add_depot(x=depot_coord[0], y=depot_coord[1])
        max_num_available = len(demands)
        cvrp_model.add_vehicle_type(capacity=capacity, num_available=max_num_available)
        clients = [
            cvrp_model.add_client(
                self.round_func(nodes_coord[idx][0]), 
                self.round_func(nodes_coord[idx][1]), 
                self.round_func(demands[idx])
            ) for idx in range(0, len(nodes_coord))
        ]
        locations = [depot] + clients
        for frm in locations:
            for to in locations:
                distance = self.get_distance(x1=(frm.x, frm.y), x2=(to.x, to.y))
                cvrp_model.add_edge(frm, to, distance=self.round_func(distance))
        res = cvrp_model.solve(stop=MaxRuntime(self.time_limit))
        
        routes = res.best.get_routes() if CP38 else res.best.routes()
        tours = [0]
        for route in routes:
            tours += route.visits()
            tours.append(0)
        return tours
        
    def solve(
        self,
        depots: Union[list, np.ndarray] = None,
        points: Union[list, np.ndarray] = None,
        demands: Union[list, np.ndarray] = None,
        capacities: Union[list, np.ndarray] = None,
        norm: str = "EUC_2D",
        normalize: bool = False,
        dtype: str = "int",
        round_func: str = "round",
        num_threads: int = 1,
        show_time: bool = False,
    ) -> np.ndarray:
        # prepare
        if dtype != "int":
            import warnings
            warnings.warn("Solver input requires data of type int.")
            dtype = "int"
        self.round_func = self.get_round_func(round_func)
        self.from_data(depots, points, demands, capacities, norm, normalize)
        start_time = time.time()

        # solve
        tours = list()
        p_shape = self.points.shape
        num_points = p_shape[0]
        if num_threads == 1:
            if show_time:
                for idx in tqdm(range(num_points), desc="Solving CVRP Using PyVRP"):
                    tours.append(self._solve(
                        depot_coord=self.depots[idx],
                        nodes_coord=self.points[idx],
                        demands=self.demands[idx],
                        capacity=self.capacities[idx]
                    ))
            else:
                for idx in range(num_points):
                    tours.append(self._solve(
                        depot_coord=self.depots[idx],
                        nodes_coord=self.points[idx],
                        demands=self.demands[idx],
                        capacity=self.capacities[idx]
                    ))
        else:
            num_tqdm = num_points // num_threads
            batch_depots = self.depots.reshape(num_tqdm, num_threads, -1)
            batch_demands = self.demands.reshape(num_tqdm, num_threads, -1)
            batch_capacities = self.capacities.reshape(num_tqdm, num_threads)
            batch_points = self.points.reshape(-1, num_threads, p_shape[-2], p_shape[-1])
            if show_time:
                for idx in tqdm(
                    range(num_points // num_threads), desc="Solving CVRP Using PyVRP"
                ):
                    with Pool(num_threads) as p1:
                        cur_tours = p1.starmap(
                            self._solve,
                            [  (batch_depots[idx][inner_idx], 
                                batch_points[idx][inner_idx], 
                                batch_demands[idx][inner_idx], 
                                batch_capacities[idx][inner_idx]) 
                                for inner_idx in range(num_threads)
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)
            else:
                for idx in range(num_points // num_threads):
                    with Pool(num_threads) as p1:
                        cur_tours = p1.starmap(
                            self._solve,
                            [  (batch_depots[idx][inner_idx], 
                                batch_points[idx][inner_idx], 
                                batch_demands[idx][inner_idx], 
                                batch_capacities[idx][inner_idx]) 
                                for inner_idx in range(num_threads)
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)

        # format
        self.read_tours(tours)
        end_time = time.time()
        if show_time:
            print(f"Use Time: {end_time - start_time}")
        return self.tours


import os
import sys
import time
import shutil
import itertools
import numpy as np
import pathlib
from tqdm import tqdm
from typing import Union
from multiprocessing import Pool
from ml4co_kit.solver.tsp import TSPLKHSolver
from ml4co_kit.evaluate.tsp.base import TSPEvaluator
from ml4co_kit.solver.tsp import TSPSolver, TSPLKHSolver, TSPConcordeSolver
import warnings

warnings.filterwarnings("ignore")


class TSPDataGenerator:
    def __init__(
        self,
        num_threads: int = 1,
        nodes_num: int = 50,
        data_type: str = "uniform",
        solver: Union[str, TSPSolver] = "lkh",
        train_samples_num: int = 128000,
        val_samples_num: int = 1280,
        test_samples_num: int = 1280,
        save_path: pathlib.Path = "data/tsp/uniform",
        filename: str = None,
        # special for gaussian
        gaussian_mean_x: float = 0.0,
        gaussian_mean_y: float = 0.0,
        gaussian_std: float = 1.0,
        # special for cluster
        cluster_nums: int = 10,
        cluster_std: float = 0.1,
        # special for regret
        regret: bool = False,
        regret_save_path: str = None,
        regret_solver: TSPSolver = None,
    ):
        """
        TSPDataGenerator
        Args:
            num_threads (int, optional):
                The number of threads to generate datasets.
            nodes_num (int, optional):
                The number of nodes.
            data_type (str, optional):
                The data type.
            solver_type (str, optional):
                The solver type.
            train_samples_num (int, optional):
                The number of training samples.
            val_samples_num (int, optional):
                The number of validation samples.
            test_samples_num (int, optional):
                The number of test samples.
            save_path (pathlib.Path, optional):
                The save path.
            filename (str, optional):
                The filename.
            gaussian_mean_x (float, optional):
                The mean of the x-coordinate in Gaussian data generation.
            gaussian_mean_y (float, optional):
                The mean of the y-coordinate in Gaussian data generation.
            gaussian_std (float, optional):
                The standard deviation in Gaussian data generation.
        """
        # record variable data
        self.num_threads = num_threads
        self.nodes_num = nodes_num
        self.data_type = data_type
        self.solver = solver
        self.train_samples_num = train_samples_num
        self.val_samples_num = val_samples_num
        self.test_samples_num = test_samples_num
        self.save_path = save_path
        self.filename = filename
        # special for gaussian
        self.gaussian_mean_x = gaussian_mean_x
        self.gaussian_mean_y = gaussian_mean_y
        self.gaussian_std = gaussian_std
        # special for cluster
        self.cluster_nums = cluster_nums
        self.cluster_std = cluster_std
        # special for regret
        self.regret = regret
        self.regret_save_path = regret_save_path
        self.regret_solver = regret_solver
        # check the input variables
        self.sample_types = ["train", "val", "test"]
        self.check_num_threads()
        self.check_data_type()
        self.check_solver()
        self.get_filename()
        self.check_regret()

    def check_num_threads(self):
        self.samples_num = 0
        for sample_type in self.sample_types:
            self.samples_num += getattr(self, f"{sample_type}_samples_num")
            if self.samples_num % self.num_threads != 0:
                message = "``samples_num`` must be divisible by the number of threads. "
                raise ValueError(message)

    def check_data_type(self):
        generate_func_dict = {
            "uniform": self.generate_uniform,
            "gaussian": self.generate_gaussian,
            "cluster": self.generate_cluster,
        }
        supported_data_type = generate_func_dict.keys()
        if self.data_type not in supported_data_type:
            message = (
                f"The input data_type ({self.data_type}) is not a valid type, "
                f"and the generator only supports {supported_data_type}."
            )
            raise ValueError(message)
        self.generate_func = generate_func_dict[self.data_type]

    def check_solver(self):
        # get solver
        if type(self.solver) == str:
            self.solver_type = self.solver
            supported_solver_dict = {"lkh": TSPLKHSolver, "concorde": TSPConcordeSolver}
            supported_solver_type = supported_solver_dict.keys()
            if self.solver_type not in supported_solver_type:
                message = (
                    f"The input solver_type ({self.solver_type}) is not a valid type, "
                    f"and the generator only supports {supported_solver_type}."
                )
                raise ValueError(message)
            self.solver = supported_solver_dict[self.solver_type]()
        else:
            self.solver: TSPSolver
            self.solver_type = self.solver.solver_type
        # check solver
        check_solver_dict = {
            "lkh": self.check_lkh,
            "concorde": self.check_concorde,
            "concorde-large": self.check_concorde,
        }
        check_func = check_solver_dict[self.solver_type]
        check_func()

    def check_lkh(self):
        # check if lkh is downloaded
        if shutil.which(self.solver.lkh_path) is None:
            self.download_lkh()
        # check again
        if shutil.which(self.solver.lkh_path) is None:
            message = (
                f"The LKH solver cannot be found in the path '{self.solver.lkh_path}'. "
                "Please make sure that you have entered the correct ``lkh_path``."
                "If you have not installed the LKH solver, "
                "please use function ``self.download_lkh()`` to download it."
                "Please also confirm whether the Conda environment of the terminal "
                "is consistent with the Python environment."
            )
            raise ValueError(message)

    def check_concorde(self):
        try:
            from ml4co_kit.solver.tsp.pyconcorde import TSPConSolver
        except:
            self.recompile_concorde()

    def recompile_concorde(self):
        concorde_path = pathlib.Path(__file__).parent.parent / "solver/tsp/pyconcorde"
        ori_dir = os.getcwd()
        os.chdir(concorde_path)
        os.system("python ./setup.py build_ext --inplace")
        os.chdir(ori_dir)

    def download_lkh(self):
        # download
        import wget
        lkh_url = "http://akira.ruc.dk/~keld/research/LKH-3/LKH-3.0.7.tgz"
        wget.download(url=lkh_url, out="LKH-3.0.7.tgz")
        # tar .tgz file
        os.system("tar xvfz LKH-3.0.7.tgz")
        # build LKH
        ori_dir = os.getcwd()
        os.chdir("LKH-3.0.7")
        os.system("make")
        # move LKH to the bin dir
        target_dir = os.path.join(sys.prefix, "bin")
        os.system(f"cp LKH {target_dir}")
        os.chdir(ori_dir)
        # delete .tgz file
        os.remove("LKH-3.0.7.tgz")
        shutil.rmtree("LKH-3.0.7")

    def get_filename(self):
        self.filename = (
            f"tsp{self.nodes_num}_{self.data_type}"
            if self.filename is None
            else self.filename
        )
        self.file_save_path = os.path.join(self.save_path, self.filename + ".txt")
        for sample_type in self.sample_types:
            setattr(
                self,
                f"{sample_type}_file_save_path",
                os.path.join(self.save_path, self.filename + f"_{sample_type}.txt"),
            )
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)

    def check_regret(self):
        if not self.regret:
            return
        if self.regret_save_path is None:
            self.regret_save_path = os.path.join(self.save_path, "regret")
        if self.regret_solver is None:
            self.regret_solver = TSPLKHSolver(lkh_max_trials=10)
        if not os.path.exists(self.regret_save_path):
            os.makedirs(self.regret_save_path)

    def generate(self):
        start_time = time.time()
        cnt = 0
        for _ in tqdm(
            range(self.samples_num // self.num_threads),
            desc=f"Solving TSP Using {self.solver_type}",
        ):
            # call generate_func to generate the points 
            batch_nodes_coord = self.generate_func()
            with Pool(self.num_threads) as p1:
                tours = p1.map(
                    self.solver.solve,
                    [batch_nodes_coord[idx] for idx in range(self.num_threads)],
                )
            
            # deal with regret
            if self.regret:
                p1.close()  # Close the pool to indicate that no more tasks will be submitted
                p1.join()  # Wait for all processes in the pool to complete
                if self.num_threads == 1:
                    self.generate_regret(tours[0], batch_nodes_coord[0], cnt)    
                else:
                    with Pool(self.num_threads) as p2:
                        p2.starmap(
                            self.generate_regret,
                            [
                                (tour, batch_nodes_coord[idx], cnt + idx)
                                for idx, tour in enumerate(tours)
                            ],
                        )
                cnt += self.num_threads
            
            # write to txt
            for idx, tour in enumerate(tours):
                tour = tour[:-1]
                if (np.sort(tour) == np.arange(self.nodes_num)).all():
                    with open(self.file_save_path, "a+") as f:
                        f.write(
                            " ".join(
                                str(x) + str(" ") + str(y)
                                for x, y in batch_nodes_coord[idx]
                            )
                        )
                        f.write(str(" ") + str("output") + str(" "))
                        f.write(str(" ").join(str(node_idx + 1) for node_idx in tour))
                        f.write(str(" ") + str(tour[0] + 1) + str(" "))
                        f.write("\n")
                    f.close()
        
        # info
        end_time = time.time() - start_time
        print(
            f"Completed generation of {self.samples_num} samples of TSP{self.nodes_num}."
        )
        print(f"Total time: {end_time/60:.1f}m")
        print(f"Average time: {end_time/self.samples_num:.1f}s")
        self.devide_file()

    def generate_regret(self, tour: np.ndarray, nodes_coord: np.ndarray, cnt: int):
        opt_tour = list(tour) + [0]
        reg_mat = self.calc_regret(nodes_coord, opt_tour)
        np.save(os.path.join(self.regret_save_path, f"{cnt}.npy"), reg_mat)

    def devide_file(self):
        with open(self.file_save_path, "r") as f:
            data = f.readlines()
        train_end_idx = self.train_samples_num
        val_end_idx = self.train_samples_num + self.val_samples_num
        train_data = data[:train_end_idx]
        val_data = data[train_end_idx:val_end_idx]
        test_data = data[val_end_idx:]
        data = [train_data, val_data, test_data]
        for sample_type, data_content in zip(self.sample_types, data):
            filename = getattr(self, f"{sample_type}_file_save_path")
            with open(filename, "w") as file:
                file.writelines(data_content)
        # deal with regret
        if self.regret:
            for root, _, file in os.walk(self.regret_save_path):
                file.sort(key=lambda x: int(x.split(".")[0]))
                for i, reg_file in enumerate(file):
                    if i < self.train_samples_num:
                        shutil.move(
                            os.path.join(root, reg_file),
                            os.path.join(self.regret_save_path, f"train_{i}.npy"),
                        )
                    elif i < self.train_samples_num + self.val_samples_num:
                        shutil.move(
                            os.path.join(root, reg_file),
                            os.path.join(
                                self.regret_save_path, f"val_{i-train_end_idx}.npy"
                            ),
                        )
                    else:
                        shutil.move(
                            os.path.join(root, reg_file),
                            os.path.join(
                                self.regret_save_path, f"test_{i-val_end_idx}.npy"
                            ),
                        )
                break

    def generate_uniform(self) -> np.ndarray:
        return np.random.random([self.num_threads, self.nodes_num, 2])

    def generate_gaussian(self) -> np.ndarray:
        return np.random.normal(
            loc=[self.gaussian_mean_x, self.gaussian_mean_y],
            scale=self.gaussian_std,
            size=(self.num_threads, self.nodes_num, 2),
        )

    def generate_cluster(self):
        nodes_coords = np.zeros([self.num_threads, self.nodes_num, 2])
        for i in range(self.num_threads):
            cluster_centers = np.random.random([self.cluster_nums, 2])
            cluster_points = []
        for center in cluster_centers:
            points = np.random.normal(
                loc=center,
                scale=self.cluster_std,
                size=(self.nodes_num // self.cluster_nums, 2),
            )
            cluster_points.append(points)
        nodes_coords[i] = np.concatenate(cluster_points, axis=0)
        return nodes_coords

    def calc_regret(self, points: np.ndarray, opt_tour: list):
        num_nodes = points.shape[0]
        reg_mat = np.zeros((num_nodes, num_nodes))
        eva = TSPEvaluator(points)
        for i, j in itertools.combinations(range(num_nodes), 2):
            tour = self.regret_solver.regret_solve(points=points, fixed_edges=(i, j))
            cost = eva.evaluate(tour)
            opt_cost = eva.evaluate(opt_tour)
            regret = (cost - opt_cost) / opt_cost
            reg_mat[i, j] = reg_mat[j, i] = regret
        return reg_mat



import torch
import numpy as np
import torch.nn.functional as F
from ml4co_kit.learning.model import BaseModel
from ml4co_kit.evaluate import TSPEvaluator
from ml4co_kit.learning.search import SearchConfigurator
from env import TSPEnv
from search import tsp_greedy, tsp_2opt
from typing import Any, Union


class TSPGNN(BaseModel):
    def __init__(
        self,
        nodes_num: int,
        # network
        network_type: str = "gnn",
        input_dim: int = 2,
        hidden_dim: int = 256,
        output_channels: int = 2,
        num_layers: int = 12,
        # env
        mode: str = "train",
        train_path: str = None,
        val_path: str = None,
        test_path: str = None,
        train_batch_size: int = 64,
        val_batch_size: int = 1,
        test_batch_size: int = 1,
        valid_samples: int = 1280,
        num_workers: int = 4,
        # training params
        lr_scheduler: str = "cosine-decay",
        learning_rate: float = 2e-4,
        weight_decay: float = 1e-4,
        # test_step
        decoding_type: str = "greedy",
        local_search_type: str = "2opt",
        **kwargs,
    ):
        # network and model
        self.net = self.get_net(network_type)
        model = self.net(
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            output_channels=output_channels,
            num_layers=num_layers,
        )

        # env
        env = TSPEnv(
            nodes_num=nodes_num,
            mode=mode,
            train_path=train_path,
            val_path=val_path,
            test_path=test_path,
            train_batch_size=train_batch_size,
            val_batch_size=val_batch_size,
            test_batch_size=test_batch_size,
            valid_samples=valid_samples,
            num_workers=num_workers,
        )

        self.search_config = SearchConfigurator(
            decoding_class={("tsp", "greedy"): tsp_greedy},
            local_search_class={("tsp", "2opt"): tsp_2opt},
        )

        # super
        super(TSPGNN, self).__init__(
            env=env,
            model=model,
            lr_scheduler=lr_scheduler,
            learning_rate=learning_rate,
            weight_decay=weight_decay,
        )

        # nodes num
        self.nodes_num = nodes_num

        # test
        self.decoding_type = decoding_type
        self.test_decoding_kwargs = kwargs
        self.local_search_type = local_search_type
        self.test_ls_kwargs = kwargs

    def shared_step(self, batch: Any, batch_idx: int, phase: str):
        """Shared step between train/val/test. To be implemented."""
        edge_index = None  # for sparse graph input
        np_edge_index = None  # for sparse graph input
        _, points, adj_matrix, ref_tour = batch
        points: torch.Tensor
        ref_tour: torch.Tensor
        if phase != "train":
            np_points = points.cpu().numpy()[0]
            np_ref_tour = ref_tour.cpu().numpy()[0]

        # forward
        x_pred = self.forward(points, edge_index)
        loss = self.loss_fn(x_pred, adj_matrix)

        # return loss if current is a training step
        if phase == "train":
            metrics = {"train/loss": loss}
            for k, v in metrics.items():
                self.log(k, v, prog_bar=True, on_epoch=True, sync_dist=True)
            return loss

        # gain heatmap
        heatmap = F.softmax(x_pred, dim=1)
        adj_mat = heatmap[:, 1]

        # decoding / solve
        if phase == "val":
            adj_mat = adj_mat.cpu().numpy()
            solved_tours = tsp_greedy(
                adj_mat=adj_mat,
                np_points=np_points,
                edge_index_np=np_edge_index,
                device=ref_tour.device,
            )
        else:
            # decode
            decoding_func = self.search_config.get_decoding_func(
                task="tsp", name=self.decoding_type
            )
            solved_tours = decoding_func(
                adj_mat=adj_mat,
                np_points=np_points,
                edge_index_np=np_edge_index,
                device=ref_tour.device,
                **self.test_decoding_kwargs,
            )

            # local_search
            local_search_func = self.search_config.get_local_search_func(
                task="tsp", name=self.local_search_type
            )
            solved_tours = local_search_func(
                np_points=np_points,
                tours=solved_tours,
                adj_mat=adj_mat,
                device=ref_tour.device,
                **self.test_ls_kwargs,
            )

        # Check the tours
        for idx in range(len(solved_tours)):
            assert sorted(solved_tours[idx][:-1]) == [i for i in range(self.nodes_num)]

        # Calculate the gap
        tsp_solver = TSPEvaluator(np_points)
        ref_cost = tsp_solver.evaluate(np_ref_tour)
        all_solved_costs = [
            tsp_solver.evaluate(solved_tours[i]) for i in range(self.parallel_sampling)
        ]
        best_solved_cost = np.min(all_solved_costs)
        gap = (best_solved_cost - ref_cost) / ref_cost * 100

        # record the better/worse/match
        better = 0.0
        match = 0.0
        worse = 0.0
        if gap < -1e-12:
            better = 1.0
        elif gap < 1e-12:
            match = 1.0
        else:
            worse = 1.0
        self.gap_list.append(gap)

        # Log the loss and gap
        metrics = {
            f"{phase}/loss": loss,
            f"{phase}/gap": gap,
            f"{phase}/better": better,
            f"{phase}/match": match,
            f"{phase}/worse": worse,
        }

        if phase == "test":
            metrics.update(
                {"test/ref_cost": ref_cost, "test/solved_cost": best_solved_cost}
            )

        for k, v in metrics.items():
            self.log(k, v, prog_bar=True, on_epoch=True, sync_dist=True)
        return metrics

    def solve(
        self,
        data: Union[np.ndarray, torch.Tensor],
        edge_index: Union[np.ndarray, torch.Tensor] = None,
        batch_size: int = 16,
        device="cpu",
    ):
        """solve function, return heatmap"""
        raise NotImplementedError

    def load_weights(self, pretrained_path: str = None):
        raise NotImplementedError

    def loss_fn(self, x_pred, adj_matrix):
        raise NotImplementedError



import os
from ml4co_kit.utils import download, extract_archive


TSP_UNIFORM_50 = [
    "dataset/tsp_uniform/tsp50_lkh_100_5.68760.txt",
    "dataset/tsp_uniform/tsp50_lkh_500_5.68759.txt",
    "dataset/tsp_uniform/tsp50_concorde_5.68759.txt",
]

TSP_UNIFORM_100 = [
    "dataset/tsp_uniform/tsp100_lkh_100_7.75653.txt",
    "dataset/tsp_uniform/tsp100_lkh_500_7.75598.txt",
    "dataset/tsp_uniform/tsp100_lkh_1k_7.75592.txt",
    "dataset/tsp_uniform/tsp100_lkh_5k_7.75588.txt",
    "dataset/tsp_uniform/tsp100_concorde_5_7.75585.txt",
]

TSP_UNIFORM_500 = [
    "dataset/tsp_uniform/tsp500_lkh_100_16.65813.txt"
    "dataset/tsp_uniform/tsp500_lkh_500_16.59910.txt",
    "dataset/tsp_uniform/tsp500_lkh_5k_16.55766.txt",
    "dataset/tsp_uniform/tsp500_lkh_10k_16.55283.txt",
    "dataset/tsp_uniform/tsp500_lkh_100k_16.54753.txt",
    "dataset/tsp_uniform/tsp500_concorde_16.54581.txt",
]

TSP_UNIFORM_1000 = [
    "dataset/tsp_uniform/tsp1000_concorde_23.11812.txt",
]

TSP_UNIFORM_10000 = [
    "dataset/tsp_uniform/tsp10000_lkh_100k_72.23920.txt",
    "dataset/tsp_uniform/tsp10000_lkh_1m_72.00978.txt",
    "dataset/tsp_uniform/tsp10000_lkh_5m_71.93551.txt",
]


class TSPUniformDataset:
    def __init__(self):
        self.url = "https://huggingface.co/datasets/ML4CO/TSPUniformDataset/resolve/main/tsp_uniform.tar.gz?download=true"
        self.md5 = "494766b3ab67105563e855c7a1f22d80"
        self.dir = "dataset/tsp_uniform"
        self.raw_data_path = "dataset/tsp_uniform.tar.gz"
        if not os.path.exists("dataset"):
            os.mkdir("dataset")
        if not os.path.exists(self.dir):
            download(filename=self.raw_data_path, url=self.url, md5=self.md5)
            extract_archive(archive_path=self.raw_data_path, extract_path=self.dir)

    @property
    def supported(self):
        supported_files = {
            50: TSP_UNIFORM_50,
            100: TSP_UNIFORM_100,
            500: TSP_UNIFORM_500,
            1000: TSP_UNIFORM_1000,
            10000: TSP_UNIFORM_10000,
        }
        return supported_files



import networkx as nx


SUPPORT_POS_TYPE_DICT = {
    "bipartite_layout": nx.bipartite_layout,
    "circular_layout": nx.circular_layout,
    "kamada_kawai_layout": nx.kamada_kawai_layout,
    "random_layout": nx.random_layout,
    "rescale_layout": nx.rescale_layout,
    "rescale_layout_dict": nx.rescale_layout_dict,
    "shell_layout": nx.shell_layout,
    "spring_layout": nx.spring_layout,
    "spectral_layout": nx.spectral_layout,
    "planar_layout": nx.planar_layout,
    "fruchterman_reingold_layout": nx.fruchterman_reingold_layout,
    "spiral_layout": nx.spiral_layout,
    "multipartite_layout": nx.multipartite_layout,
}

SUPPORT_POS_TYPE = [
    "bipartite_layout",
    "circular_layout",
    "kamada_kawai_layout",
    "random_layout",
    "rescale_layout",
    "rescale_layout_dict",
    "shell_layout",
    "spring_layout",
    "spectral_layout",
    "planar_layout",
    "fruchterman_reingold_layout",
    "spiral_layout",
    "multipartite_layout",
]


def get_pos_layer(pos_type: str):
    if pos_type not in SUPPORT_POS_TYPE:
        raise ValueError(f"unvalid pos type, only supports {SUPPORT_POS_TYPE}")
    return SUPPORT_POS_TYPE_DICT[pos_type]
